"""Define extension for Nifti files that summarizes meta data from source DICOMs
"""
from __future__ import print_function

from functools import cached_property
from typing import Optional, Sequence, Tuple, Union
import sys, re, json, warnings
from copy import deepcopy
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

import numpy as np
import nibabel as nb
from nibabel.nifti1 import Nifti1Extension

from .globals import SORT_GUESSES
from .utils import iteritems, unicode_str, PY2
from .info import __version__


dcm_meta_ecode = 0


_meta_version = 0.7


_req_base_keys_map= {
    0.5 : set((
        'dcmmeta_affine',
        'dcmmeta_slice_dim',
        'dcmmeta_shape',
        'dcmmeta_version',
        'global',
    )),
    0.6 : set((
        'dcmmeta_affine',
        'dcmmeta_reorient_transform',
        'dcmmeta_slice_dim',
        'dcmmeta_shape',
        'dcmmeta_version',
        'global',
    )),
    0.7 : set((
        'dcmmeta_affine',
        'dcmmeta_reorient_transform',
        'dcmmeta_slice_dim',
        'dcmmeta_time_dim',
        'dcmmeta_vector_dim',
        'dcmmeta_shape',
        'dcmmeta_version',
        'extract_version',
        'global',
    )),
}
'''Minimum required keys in the base dictionaty to be considered valid'''


def is_constant(sequence, period=None):
    '''Returns true if all elements in (each period of) the sequence are equal.

    Parameters
    ----------
    sequence : sequence
        The sequence of elements to check.

    period : int
        If not None then each subsequence of that length is checked.
    '''
    if period is None:
        return all(val == sequence[0] for val in sequence)
    else:
        if period <= 1:
            raise ValueError('The period must be greater than one')
        seq_len = len(sequence)
        if seq_len % period != 0:
            raise ValueError('The sequence length is not evenly divisible by '
                             'the period length.')
        for period_idx in range(seq_len // period):
            start_idx = period_idx * period
            end_idx = start_idx + period
            if not all(val == sequence[start_idx]
                       for val in sequence[start_idx:end_idx]):
                return False

    return True


def is_repeating(sequence, period):
    '''Returns true if the elements in the sequence repeat with the given
    period.

    Parameters
    ----------
    sequence : sequence
        The sequence of elements to check.

    period : int
        The period over which the elements should repeat.
    '''
    seq_len = len(sequence)
    if period <= 1 or period >= seq_len:
        raise ValueError('The period must be greater than one and less than '
                         'the length of the sequence')
    if seq_len % period != 0:
        raise ValueError('The sequence length is not evenly divisible by the '
                         'period length.')

    for period_idx in range(1, seq_len // period):
        start_idx = period_idx * period
        end_idx = start_idx + period
        if sequence[start_idx:end_idx] != sequence[:period]:
            return False

    return True


class InvalidExtensionError(Exception):
    def __init__(self, msg):
        '''Exception denoting than a DcmMetaExtension is invalid.'''
        self.msg = msg

    def __str__(self):
        return 'The extension is not valid: %s' % self.msg


class DcmMetaExtension(Nifti1Extension):
    '''Nifti extension for storing a summary of the meta data from the source
    DICOM files.
    '''

    @classmethod
    def make_empty(
        klass, 
        shape: Tuple[int, ...], 
        affine: np.ndarray, 
        reorient_transform: Optional[np.ndarray] = None,
        slice_dim: Optional[int] = None,
        time_dim: Optional[str] = None,
        vector_dim: Optional[str] = None,
        extract_version: Optional[str] = None,
    ) -> "DcmMetaExtension":
        '''Make an empty DcmMetaExtension

        Parameters
        ----------
        shape
            The shape of the data associated with this extension.

        affine
            The RAS affine for the data associated with this extension.

        reorient_transform
            The transformation matrix representing any reorientation of the
            data array.

        slice_dim
            The index of the slice dimension for the data associated with this
            extension

        time_dim
            The attribute used to order the time dimension

        vector_dim
            The attribute used to order the vector dimension
        '''
        result = klass(dcm_meta_ecode, '{}')
        result._content['global'] = {}
        result._content['global']['const'] = {}
        result._content['global']['slices'] = {}
        if len(shape) > 3 and shape[3] != 1:
            result._content['time'] = {}
            result._content['time']['samples'] = {}
            result._content['time']['slices'] = {}
        else:
            time_dim = None
        if len(shape) > 4:
            result._content['vector'] = {}
            result._content['vector']['samples'] = {}
            result._content['vector']['slices'] = {}
        else:
            vector_dim = None
        result._content['dcmmeta_shape'] = []
        result.shape = shape
        result.affine = affine
        result.reorient_transform = reorient_transform
        result.slice_dim = slice_dim
        result.time_dim = time_dim
        result.vector_dim = vector_dim
        result.version = _meta_version
        result.extract_version = __version__ if extract_version is None else extract_version
        return result

    @classmethod
    def from_sequence(
        klass, 
        seq: Sequence["DcmMetaExtension"], 
        dim: int, 
        affine: Optional[np.ndarray] = None, 
        slice_dim: Optional[int] = None,
        time_dim: Optional[str] = None,
        vector_dim: Optional[str] = None,
    ) -> "DcmMetaExtension":
        '''Concatenate sequence `seq` of extensions along dimension `dim` 

        Parameters
        ----------
        seq
            The sequence of DcmMetaExtension objects to merge

        dim
            The dimension to merge the extensions along.

        affine
            The affine to use in the resulting extension. If None, the affine
            from the first extension in `seq` will be used.

        slice_dim
            The slice dimension to use in the resulting extension. If None, the
            slice dimension from the first extension in `seq` will be used.

        time_dim
            The attribute used to order the time dimension

        vector_dim
            The attribute used to order the vector dimension
        '''
        if not 0 <= dim < 5:
            raise ValueError("The argument 'dim' must be in the range [0, 5).")
        # Determine the output shape and create empty extension
        n_inputs = len(seq)
        first_input = seq[0]
        input_shape = first_input.shape
        if len(input_shape) > dim and input_shape[dim] != 1:
            raise ValueError("The dim must be singular or not exist for the "
                             "inputs.")
        output_shape = list(input_shape)
        while len(output_shape) <= dim:
            output_shape.append(1)
        output_shape[dim] = n_inputs
        output_shape = tuple(output_shape)
        if affine is None:
            affine = first_input.affine
        if slice_dim is None:
            slice_dim = first_input.slice_dim
        if time_dim is None:
            time_dim = first_input.time_dim
        if vector_dim is None:
            vector_dim = first_input.vector_dim
        result = klass.make_empty(
            output_shape,
            affine,
            None,
            slice_dim,
            time_dim,
            vector_dim,
            first_input.extract_version,
        )
        # Need to initialize the result with the first extension in 'seq'
        result_slc_norm = result.slice_normal
        first_slc_norm = first_input.slice_normal
        use_slices = (not result_slc_norm is None and
                      not first_slc_norm is None and
                      np.allclose(result_slc_norm, first_slc_norm))
        for classes in first_input.valid_classes:
            if classes[1] == 'slices' and not use_slices:
                continue
            result._content[classes[0]][classes[1]] = \
                deepcopy(first_input.get_class_dict(classes))
        #Adjust the shape to what the extension actually contains
        shape = list(result.shape)
        shape[dim] = 1
        result.shape = shape
        #Initialize reorient transform
        reorient_transform = first_input.reorient_transform
        #Add the other extensions, updating the shape as we go
        for input_ext in seq[1:]:
            #If the affines or reorient_transforms don't match, we set the
            #reorient_transform to None as we can not reliably use it to update
            #directional meta data
            if ((reorient_transform is None or
                 input_ext.reorient_transform is None) or
                not (np.allclose(input_ext.affine, affine) or
                     np.allclose(input_ext.reorient_transform,
                                 reorient_transform)
                    )
               ):
                reorient_transform = None
            result._insert(dim, input_ext)
            shape[dim] += 1
            result.shape = shape
        #Set the reorient transform
        result.reorient_transform = reorient_transform
        #Try simplifying any keys in global slices
        for key in list(result.get_class_dict(('global', 'slices'))):
            result._simplify(key)
        return result

    def guess_extra_dims(self) -> None:
        extra_shape = self.shape[3:]
        if extra_shape:
            if extra_shape[0] != 1 and self.time_dim is None:
                meta = self.get_class_dict(("time", "samples"))
                meta = {k : v for k, v in meta.items() if len(v) == len(set(v))}
                for key in SORT_GUESSES:
                    if key in meta:
                        self.time_dim = key
                        break
                else:
                    meta = self.get_class_dict(("global", "slices"))
                    meta = {k : v for k, v in meta.items() if len(v) == len(set(v))}
                    for key in SORT_GUESSES:
                        if key in meta:
                            self.time_dim = key
                            break
            if len(extra_shape) == 2 and extra_shape[1] != 1 and self.vector_dim is None:
                pass # TODO

    @property
    def shape(self):
        '''The shape of the data associated with the meta data. Defines the
        number of values for the meta data classifications.'''
        return tuple(self._content['dcmmeta_shape'])

    @shape.setter
    def shape(self, value):
        if not (3 <= len(value) < 6):
            raise ValueError("The shape must have a length between three and "
                             "six")
        # TODO: If num slice / time / vector change we should remove corresponding
        #       meta data
        self._content['dcmmeta_shape'][:] = value

    @property
    def affine(self):
        '''The affine associated with the meta data. If this differs from the
        image affine, the per-slice meta data will not be used. '''
        return np.array(self._content['dcmmeta_affine'])

    @affine.setter
    def affine(self, value):
        if value.shape != (4, 4):
            raise ValueError("Invalid shape for affine")
        self._content['dcmmeta_affine'] = value.tolist()

    @property
    def reorient_transform(self):
        '''The transformation due to reorientation of the data array. Can be
        used to update directional DICOM meta data (after converting to RAS if
        needed) into the same space as the affine.'''
        if self.version < 0.6:
            return None
        if self._content['dcmmeta_reorient_transform'] is None:
            return None
        return np.array(self._content['dcmmeta_reorient_transform'])

    @reorient_transform.setter
    def reorient_transform(self, value):
        if not value is None and value.shape != (4, 4):
            raise ValueError("The reorient_transform must be none or (4,4) "
            "array")
        if value is None:
            self._content['dcmmeta_reorient_transform'] = None
        else:
            self._content['dcmmeta_reorient_transform'] = value.tolist()

    @property
    def slice_dim(self) -> Optional[int]:
        '''The index of the slice dimension associated with the per-slice meta data.'''
        return self._content['dcmmeta_slice_dim']

    @slice_dim.setter
    def slice_dim(self, value: Optional[int]):
        if value is not None and not (0 <= value < 3):
            raise ValueError("The slice dimension must be between zero and "
                             "two")
        self._content['dcmmeta_slice_dim'] = value
    
    @property
    def time_dim(self) -> Optional[str]:
        '''The meta data key used to order the 'time' dimension'''
        return self._content.get('dcmmeta_time_dim')
    
    @time_dim.setter
    def time_dim(self, value: Optional[str]) -> None:
        if value is not None:
            # TODO: Need better validation here, maybe wait for metasum work...
            if value in self.get_class_dict(("global", "const")) or value not in self.get_keys():
                raise ValueError("Invalid attribute for indexing: {value}")
        self._content['dcmmeta_time_dim'] = value

    @property
    def vector_dim(self) -> Optional[str]:
        '''The meta data key used to order the 'vector' dimension'''
        return self._content.get('dcmmeta_vector_dim')

    @vector_dim.setter
    def vector_dim(self, value: Optional[str]) -> None:
        if value is not None:
            # TODO: Need better validation here, maybe wait for metasum work...
            if value in self.get_class_dict(("global", "const")) or value not in self.get_keys():
                raise ValueError("Invalid attribute for indexing: {value}")
        self._content['dcmmeta_vector_dim'] = value

    @property
    def version(self):
        '''The version of the meta data extension.'''
        return self._content['dcmmeta_version']

    @version.setter
    def version(self, value):
        '''Set the version of the meta data extension.'''
        self._content['dcmmeta_version'] = value

    @property
    def extract_version(self) -> Optional[str]:
        return self._content["extract_version"]
    
    @extract_version.setter
    def extract_version(self, value: Optional[str]) -> None:
        self._content["extract_version"] = value

    @property
    def slice_normal(self):
        '''The slice normal associated with the per-slice meta data.'''
        slice_dim = self.slice_dim
        if slice_dim is None:
            return None
        return np.array(self.affine[slice_dim][:3])

    @property
    def n_slices(self):
        '''The number of slices associated with the per-slice meta data.'''
        slice_dim = self.slice_dim
        if slice_dim is None:
            return None
        return self.shape[slice_dim]

    # TODO: Can we provide better aliases for these, like:
    # global, per_time, per_time_slice, per_vec, per_vec_slice, per_slice
    classifications = (('global', 'const'), 
                       ('global', 'slices'),
                       ('time', 'samples'),
                       ('time', 'slices'),
                       ('vector', 'samples'),
                       ('vector', 'slices'),
                      )
    '''The classifications used to separate meta data based on if and how the
    values repeat. Each class is a tuple with a base class and a sub class.'''

    def get_valid_classes(self):
        '''Return the meta data classifications that are valid for this
        extension.

        Returns
        -------
        valid_classes : tuple
            The classifications that are valid for this extension (based on its
            shape).

        '''
        shape = self.shape
        n_dims = len(shape)
        if n_dims == 3:
            return self.classifications[:2]
        elif n_dims == 4:
            return self.classifications[:4]
        elif n_dims == 5:
            if shape[3] != 1:
                return self.classifications
            else:
                return self.classifications[:2] + self.classifications[-2:]
        else:
            raise ValueError("There must be 3 to 5 dimensions.")
        
    valid_classes = cached_property(get_valid_classes)

    def get_multiplicity(self, classification):
        '''Get the number of meta data values for all meta data of the provided
        classification.

        Parameters
        ----------
        classification : tuple
            The meta data classification.

        Returns
        -------
        multiplicity : int
            The number of values for any meta data of the provided
            `classification`.
        '''
        if not classification in self.valid_classes:
            raise ValueError("Invalid classification: %s" % classification)

        base, sub = classification
        shape = self.shape
        n_vals = 1
        if sub == 'slices':
            n_vals = self.n_slices
            if n_vals is None:
                return 0
            if base == 'vector':
                n_vals *= shape[3]
            elif base == 'global':
                for dim_size in shape[3:]:
                    n_vals *= dim_size
        elif sub == 'samples':
            if base == 'time':
                n_vals = shape[3]
                if len(shape) == 5:
                    n_vals *= shape[4]
            elif base == 'vector':
                n_vals = shape[4]

        return n_vals

    def check_valid(self):
        '''Check if the extension is valid.

        Raises
        ------
        InvalidExtensionError
            The extension is missing required meta data or classifications, or
            some element(s) have the wrong number of values for their
            classification.
        '''
        #Check for the required base keys in the json data
        if not _req_base_keys_map[self.version] <= set(self._content):
            raise InvalidExtensionError('Missing one or more required keys')

        #Check the orientation/shape/version
        if self.affine.shape != (4, 4):
            raise InvalidExtensionError('Affine has incorrect shape')
        slice_dim = self.slice_dim
        if slice_dim is not None:
            if not (0 <= slice_dim < 3):
                raise InvalidExtensionError('Slice dimension is not valid')
        if not (3 <= len(self.shape) < 6):
            raise InvalidExtensionError('Shape is not valid')

        #Check all required meta dictionaries, make sure values have correct
        #multiplicity
        valid_classes = self.valid_classes
        for classes in valid_classes:
            if not classes[0] in self._content:
                raise InvalidExtensionError('Missing required base '
                                            'classification %s' % classes[0])
            if not classes[1] in self._content[classes[0]]:
                raise InvalidExtensionError(('Missing required sub '
                                             'classification %s in base '
                                             'classification %s') % classes)
            cls_meta = self.get_class_dict(classes)
            cls_mult = self.get_multiplicity(classes)
            if cls_mult == 0 and len(cls_meta) != 0:
                raise InvalidExtensionError('Slice dim is None but per-slice '
                                            'meta data is present')
            elif cls_mult > 1:
                for key, vals in iteritems(cls_meta):
                    n_vals = len(vals)
                    if n_vals != cls_mult:
                        msg = (('Incorrect number of values for key %s with '
                                'classification %s, expected %d found %d') %
                               (key, classes, cls_mult, n_vals)
                              )
                        raise InvalidExtensionError(msg)

        #Check that all keys are uniquely classified
        for classes in valid_classes:
            for other_classes in valid_classes:
                if classes == other_classes:
                    continue
                intersect = (set(self.get_class_dict(classes)) &
                             set(self.get_class_dict(other_classes))
                            )
                if len(intersect) != 0:
                    raise InvalidExtensionError("One or more keys have "
                                                "multiple classifications")

    def get_keys(self):
        '''Get a list of all the meta data keys that are available.'''
        keys = []
        for base_class, sub_class in self.valid_classes:
            keys += self._content[base_class][sub_class].keys()
        return keys

    def get_classification(self, key):
        '''Get the classification for the given `key`.

        Parameters
        ----------
        key : str
            The meta data key.

        Returns
        -------
        classification : tuple or None
            The classification tuple for the provided key or None if the key is
            not found.

        '''
        for base_class, sub_class in self.valid_classes:
            if key in self._content[base_class][sub_class]:
                    return (base_class, sub_class)

        return None

    def get_class_dict(self, classification):
        '''Get the dictionary for the given classification.

        Parameters
        ----------
        classification : tuple
            The meta data classification.

        Returns
        -------
        meta_dict : dict
            The dictionary for the provided classification.
        '''
        base, sub = classification
        return self._content[base][sub]

    def get_values(self, key):
        '''Get all values for the provided key.

        Parameters
        ----------
        key : str
            The meta data key.

        Returns
        -------
        values
             The value or values for the given key. The number of values
             returned depends on the classification (see 'get_multiplicity').
        '''
        classification = self.get_classification(key)
        if classification is None:
            return None
        return self.get_class_dict(classification)[key]

    def get_values_and_class(self, key):
        '''Get the values and the classification for the provided key.

        Parameters
        ----------
        key : str
            The meta data key.

        Returns
        -------
        vals_and_class : tuple
            None for both the value and classification if the key is not found.

        '''
        classification = self.get_classification(key)
        if classification is None:
            return (None, None)
        return (self.get_class_dict(classification)[key], classification)

    def filter_meta(self, filter_func):
        '''Filter the meta data.

        Parameters
        ----------
        filter_func : callable
            Must take a key and values as parameters and return True if they
            should be filtered out.

        '''
        for classes in self.valid_classes:
            filtered = []
            curr_dict = self.get_class_dict(classes)
            for key, values in iteritems(curr_dict):
                if filter_func(key, values):
                    filtered.append(key)
            for key in filtered:
                del curr_dict[key]

    def clear_slice_meta(self):
        '''Clear all meta data that is per slice.'''
        for base_class, sub_class in self.valid_classes:
            if sub_class == 'slices':
                self.get_class_dict((base_class, sub_class)).clear()

    def get_subset(self, dim, idx):
        '''Get a DcmMetaExtension containing a subset of the meta data.

        Parameters
        ----------
        dim : int
            The dimension we are taking the subset along.

        idx : int
            The position on the dimension `dim` for the subset.

        Returns
        -------
        result : DcmMetaExtension
            A new DcmMetaExtension corresponding to the subset.

        '''
        if not 0 <= dim < 5:
            raise ValueError("The argument 'dim' must be in the range [0, 5).")
        shape = self.shape
        valid_classes = self.valid_classes
        #Make an empty extension for the result
        result_shape = list(shape)
        result_shape[dim] = 1
        while result_shape[-1] == 1 and len(result_shape) > 3:
            result_shape = result_shape[:-1]
        result_shape = tuple(result_shape)
        result = self.make_empty(result_shape,
                                 self.affine,
                                 self.reorient_transform,
                                 self.slice_dim,
                                 self.time_dim,
                                 self.vector_dim,
                                )
        for src_class in valid_classes:
            #Constants remain constant
            if src_class == ('global', 'const'):
                for key, val in iteritems(self.get_class_dict(src_class)):
                    result.get_class_dict(src_class)[key] = deepcopy(val)
                continue
            if dim == self.slice_dim:
                if src_class[1] != 'slices':
                    for key, vals in iteritems(self.get_class_dict(src_class)):
                        result.get_class_dict(src_class)[key] = deepcopy(vals)
                else:
                    result._copy_slice(self, src_class, idx)
            elif dim < 3:
                for key, vals in iteritems(self.get_class_dict(src_class)):
                    result.get_class_dict(src_class)[key] = deepcopy(vals)
            elif dim == 3:
                result._copy_sample(self, src_class, 'time', idx)
            else:
                result._copy_sample(self, src_class, 'vector', idx)
        return result

    def to_json(self):
        '''Return the extension encoded as a JSON string.'''
        self.check_valid()
        return json.dumps(self._content, indent=4)

    @classmethod
    def from_json(klass, json_str):
        '''Create an extension from the JSON string representation.'''
        result = klass(dcm_meta_ecode, json_str)
        result.check_valid()
        return result

    @classmethod
    def from_runtime_repr(klass, runtime_repr):
        '''Create an extension from the Python runtime representation (nested
        dictionaries).
        '''
        result = klass(dcm_meta_ecode, '{}')
        result._content = runtime_repr
        result.check_valid()
        return result

    def __str__(self) -> str:
        return json.dumps(self._content, indent=4)

    def __eq__(self, other):
        if not np.allclose(self.affine, other.affine):
            return False
        if self.shape != other.shape:
            return False
        if self.slice_dim != other.slice_dim:
            return False
        if self.version != other.version:
            return False
        for classes in self.valid_classes:
            if (dict(self.get_class_dict(classes)) !=
               dict(other.get_class_dict(classes))):
                return False

        return True

    def _unmangle(self, value):
        '''Go from extension data to runtime representation.'''
        if not isinstance(value, unicode_str):
            value = value.decode('utf-8')
        #Its not possible to preserve order while loading with python 2.6
        kwargs = {}
        if sys.version_info >= (2, 7):
            kwargs['object_pairs_hook'] = OrderedDict
        return json.loads(value, **kwargs)

    def _mangle(self, value):
        '''Go from runtime representation to extension data.'''
        res = json.dumps(value, indent=4)
        # Python 2 leaves some trailing white-space in the JSON output while
        # python 3 does not. We strip it so output is binary identical across 
        # versions
        if PY2:
            res = re.sub('[ \t]+$', '', res, 0, re.M)
        return res.encode('utf-8')

    _const_tests = {('global', 'slices') : (('global', 'const'),
                                            ('vector', 'samples'),
                                            ('time', 'samples')
                                           ),
                    ('vector', 'slices') : (('global', 'const'),
                                            ('time', 'samples')
                                           ),
                    ('time', 'slices') : (('global', 'const'),
                                         ),
                    ('time', 'samples') : (('global', 'const'),
                                           ('vector', 'samples'),
                                          ),
                    ('vector', 'samples') : (('global', 'const'),)
                   }
    '''Classification mapping showing possible reductions in multiplicity for
    values that are constant with some period.'''

    def _get_const_period(self, src_cls, dest_cls):
        '''Get the period over which we test for const-ness with for the
        given classification change.'''
        if dest_cls == ('global', 'const'):
            return None
        elif src_cls == ('global', 'slices'):
            return int(self.get_multiplicity(src_cls) // self.get_multiplicity(dest_cls))
        elif src_cls == ('vector', 'slices'): #implies dest_cls == ('time', 'samples'):
            return  self.n_slices
        elif src_cls == ('time', 'samples'): #implies dest_cls == ('vector', 'samples')
            return self.shape[3]
        assert False #Should take one of the above branches

    _repeat_tests = {('global', 'slices') : (('time', 'slices'),
                                             ('vector', 'slices')
                                            ),
                     ('vector', 'slices') : (('time', 'slices'),),
                    }
    '''Classification mapping showing possible reductions in multiplicity for
    values that are repeating with some period.'''

    def _simplify(self, key):
        '''Try to simplify (reduce the multiplicity) of a single meta data
        element by changing its classification. Return True if the
        classification is changed, otherwise False.

        Looks for values that are constant or repeating with some pattern.
        Constant elements with a value of None will be deleted.
        '''
        values, curr_class = self.get_values_and_class(key)

        #If the class is global const then just delete it if the value is None
        if curr_class == ('global', 'const'):
            if values is None:
                del self.get_class_dict(curr_class)[key]
                return True
            return False

        #Test if the values are constant with some period
        dests = self._const_tests[curr_class]
        for dest_cls in dests:
            if dest_cls[0] in self._content:
                period = self._get_const_period(curr_class, dest_cls)
                #If the period is one, the two classifications have the
                #same multiplicity so we are dealing with a degenerate
                #case (i.e. single slice data). Just change the
                #classification to the "simpler" one in this case
                if period == 1 or is_constant(values, period):
                    if period is None:
                        self.get_class_dict(dest_cls)[key] = \
                            values[0]
                    else:
                        self.get_class_dict(dest_cls)[key] = \
                            values[::period]
                    break
        else: #Otherwise test if values are repeating with some period
            if curr_class in self._repeat_tests:
                for dest_cls in self._repeat_tests[curr_class]:
                    if dest_cls[0] in self._content:
                        dest_mult = self.get_multiplicity(dest_cls)
                        if is_repeating(values, dest_mult):
                            self.get_class_dict(dest_cls)[key] = \
                                values[:dest_mult]
                            break
                else: #Can't simplify
                    return False
            else:
                return False

        del self.get_class_dict(curr_class)[key]
        return True

    _preserving_changes = {None : (('global', 'const'),
                                   ('vector', 'samples'),
                                   ('time', 'samples'),
                                   ('time', 'slices'),
                                   ('vector', 'slices'),
                                   ('global', 'slices'),
                                  ),
                           ('global', 'const') : (('vector', 'samples'),
                                                  ('time', 'samples'),
                                                  ('time', 'slices'),
                                                  ('vector', 'slices'),
                                                  ('global', 'slices'),
                                                 ),
                           ('vector', 'samples') : (('time', 'samples'),
                                                    ('global', 'slices'),
                                                   ),
                           ('time', 'samples') : (('global', 'slices'),
                                                 ),
                           ('time', 'slices') : (('vector', 'slices'),
                                                 ('global', 'slices'),
                                                ),
                           ('vector', 'slices') : (('global', 'slices'),
                                                  ),
                           ('global', 'slices') : tuple(),
                          }
    '''Classification mapping showing allowed changes when increasing the
    multiplicity.'''

    def _get_changed_class(self, key, new_class, slice_dim=None):
        '''Get an array of values corresponding to a single meta data
        element with its classification changed by increasing its
        multiplicity. This will preserve all the meta data and allow easier
        merging of values with different classifications.'''
        values, curr_class = self.get_values_and_class(key)
        if curr_class == new_class:
            return values

        if not new_class in self._preserving_changes[curr_class]:
            raise ValueError("Classification change would lose data.")

        if curr_class is None:
            curr_mult = 1
            per_slice = False
        else:
            curr_mult = self.get_multiplicity(curr_class)
            per_slice = curr_class[1] == 'slices'
        if new_class in self.valid_classes:
            new_mult = self.get_multiplicity(new_class)
            #Only way we get 0 for mult is if slice dim is undefined
            if new_mult == 0:
                new_mult = self.shape[slice_dim]
        else:
            new_mult = 1
        mult_fact = int(new_mult // curr_mult)
        if curr_mult == 1:
            values = [values]


        if per_slice:
            result = values * mult_fact
        else:
            result = []
            for value in values:
                result.extend([deepcopy(value)] * mult_fact)

        if new_class == ('global', 'const'):
            result = result[0]

        return result

    def _change_class(self, key, new_class):
        '''Change the classification of the meta data element in place. See
        _get_changed_class.'''
        values, curr_class = self.get_values_and_class(key)
        if curr_class == new_class:
            return

        self.get_class_dict(new_class)[key] = self._get_changed_class(key,
                                                                      new_class)

        if not curr_class is None:
            del self.get_class_dict(curr_class)[key]

    def _copy_slice(self, other, src_class, idx):
        '''Get a copy of the meta data from the 'other' instance with
        classification 'src_class', corresponding to the slice with index
        'idx'.'''
        if src_class[0] == 'global':
            for classes in (('time', 'samples'),
                            ('vector', 'samples'),
                            ('global', 'const')):
                if classes in self.valid_classes:
                    dest_class = classes
                    break
        elif src_class[0] == 'vector':
            for classes in (('time', 'samples'),
                            ('global', 'const')):
                if classes in self.valid_classes:
                    dest_class = classes
                    break
        else:
            dest_class = ('global', 'const')

        src_dict = other.get_class_dict(src_class)
        dest_dict = self.get_class_dict(dest_class)
        dest_mult = self.get_multiplicity(dest_class)
        stride = other.n_slices
        for key, vals in iteritems(src_dict):
            subset_vals = vals[idx::stride]

            if len(subset_vals) < dest_mult:
                full_vals = []
                for val_idx in range(dest_mult // len(subset_vals)):
                    full_vals += deepcopy(subset_vals)
                subset_vals = full_vals
            if len(subset_vals) == 1:
                subset_vals = subset_vals[0]
            dest_dict[key] = deepcopy(subset_vals)
            self._simplify(key)

    def _global_slice_subset(self, key, sample_base, idx):
        '''Get a subset of the meta data values with the classificaion
        ('global', 'slices') corresponding to a single sample along the
        time or vector dimension (as specified by 'sample_base' and 'idx').
        '''
        n_slices = self.n_slices
        shape = self.shape
        src_dict = self.get_class_dict(('global', 'slices'))
        if sample_base == 'vector':
            slices_per_vec = n_slices * shape[3]
            start_idx = idx * slices_per_vec
            end_idx = start_idx + slices_per_vec
            return src_dict[key][start_idx:end_idx]
        else:
            if not ('vector', 'samples') in self.valid_classes:
                start_idx = idx * n_slices
                end_idx = start_idx + n_slices
                return src_dict[key][start_idx:end_idx]
            else:
                result = []
                slices_per_vec = n_slices * shape[3]
                for vec_idx in range(shape[4]):
                    start_idx = (vec_idx * slices_per_vec) + (idx * n_slices)
                    end_idx = start_idx + n_slices
                    result.extend(src_dict[key][start_idx:end_idx])
                return result

    def _copy_sample(self, other, src_class, sample_base, idx):
        '''Get a copy of meta data from 'other' instance with classification
        'src_class', corresponding to one sample along the time or vector
        dimension.'''
        assert src_class != ('global', 'const')
        src_dict = other.get_class_dict(src_class)
        if src_class[1] == 'samples':
            #If we are indexing on the same dim as the src_class we need to
            #change the classification
            if src_class[0] == sample_base:
                #Time samples may become vector samples, otherwise const
                best_dest = None
                for dest_cls in (('vector', 'samples'),
                                 ('global', 'const')):
                    if (dest_cls != src_class and
                        dest_cls in self.valid_classes
                       ):
                        best_dest = dest_cls
                        break

                dest_mult = self.get_multiplicity(dest_cls)
                if dest_mult == 1:
                    for key, vals in iteritems(src_dict):
                        self.get_class_dict(dest_cls)[key] = \
                            deepcopy(vals[idx])
                else: #We must be doing time samples -> vector samples
                    stride = other.shape[3]
                    for key, vals in iteritems(src_dict):
                        self.get_class_dict(dest_cls)[key] = \
                            deepcopy(vals[idx::stride])
                    for key in src_dict.keys():
                        self._simplify(key)

            else: #Otherwise classification does not change
                #The multiplicity will change for time samples if splitting
                #vector dimension
                if src_class == ('time', 'samples'):
                    dest_mult = self.get_multiplicity(src_class)
                    start_idx = idx * dest_mult
                    end_idx = start_idx + dest_mult
                    for key, vals in iteritems(src_dict):
                        self.get_class_dict(src_class)[key] = \
                            deepcopy(vals[start_idx:end_idx])
                        self._simplify(key)
                else: #Otherwise multiplicity is unchanged
                    for key, vals in iteritems(src_dict):
                        self.get_class_dict(src_class)[key] = deepcopy(vals)
        else: #The src_class is per slice
            if src_class[0] == sample_base:
                best_dest = None
                for dest_class in self._preserving_changes[src_class]:
                    if dest_class in self.valid_classes:
                        best_dest = dest_class
                        break
                for key, vals in iteritems(src_dict):
                    self.get_class_dict(best_dest)[key] = deepcopy(vals)
            elif src_class[0] != 'global':
                if sample_base == 'time':
                    #Take a subset of vector slices
                    n_slices = self.n_slices
                    start_idx = idx * n_slices
                    end_idx = start_idx + n_slices
                    for key, vals in iteritems(src_dict):
                        self.get_class_dict(src_class)[key] = \
                            deepcopy(vals[start_idx:end_idx])
                        self._simplify(key)
                else:
                    #Time slices are unchanged
                    for key, vals in iteritems(src_dict):
                        self.get_class_dict(src_class)[key] = deepcopy(vals)
            else:
                #Take a subset of global slices
                for key, vals in iteritems(src_dict):
                    subset_vals = \
                        other._global_slice_subset(key, sample_base, idx)
                    self.get_class_dict(src_class)[key] = deepcopy(subset_vals)
                    self._simplify(key)

    def _insert(self, dim, other):
        self_slc_norm = self.slice_normal
        other_slc_norm = other.slice_normal

        #If we are not using slice meta data, temporarily remove it from the
        #other dcmmeta object
        use_slices = (not self_slc_norm is None and
                      not other_slc_norm is None and
                      np.allclose(self_slc_norm, other_slc_norm))
        other_slc_meta = {}
        if not use_slices:
            for classes in other.valid_classes:
                if classes[1] == 'slices':
                    other_slc_meta[classes] = other.get_class_dict(classes)
                    other._content[classes[0]][classes[1]] = {}
        missing_keys = list(set(self.get_keys()) - set(other.get_keys()))
        for other_classes in other.valid_classes:
            other_keys = list(other.get_class_dict(other_classes).keys())

            #Treat missing keys as if they were in global const and have a value
            #of None
            if other_classes == ('global', 'const'):
                other_keys += missing_keys

            #When possible, reclassify our meta data so it matches the other
            #classification
            for key in other_keys:
                local_classes = self.get_classification(key)
                if local_classes != other_classes:
                    local_allow = self._preserving_changes[local_classes]
                    other_allow = self._preserving_changes[other_classes]

                    if other_classes in local_allow:
                        self._change_class(key, other_classes)
                    elif not local_classes in other_allow:
                        best_dest = None
                        for dest_class in local_allow:
                            if (dest_class[0] in self._content and
                               dest_class in other_allow):
                                best_dest = dest_class
                                break
                        self._change_class(key, best_dest)

            #Insert new meta data and further reclassify as necessary
            for key in other_keys:
                if dim == self.slice_dim:
                    self._insert_slice(key, other)
                elif dim < 3:
                    self._insert_non_slice(key, other)
                elif dim == 3:
                    self._insert_sample(key, other, 'time')
                elif dim == 4:
                    self._insert_sample(key, other, 'vector')

        #Restore per slice meta if needed
        if not use_slices:
            for classes in other.valid_classes:
                if classes[1] == 'slices':
                    other._content[classes[0]][classes[1]] = \
                        other_slc_meta[classes]

    def _insert_slice(self, key, other):
        local_vals, classes = self.get_values_and_class(key)
        other_vals = other._get_changed_class(key, classes, self.slice_dim)


        #Handle some common / simple insertions with special cases
        if classes == ('global', 'const'):
            if local_vals != other_vals:
                for dest_base in ('time', 'vector', 'global'):
                    if dest_base in self._content:
                        self._change_class(key, (dest_base, 'slices'))
                        other_vals = other._get_changed_class(key,
                                                              (dest_base,
                                                               'slices'),
                                                               self.slice_dim
                                                             )
                        self.get_values(key).extend(other_vals)
                        break
        elif classes == ('time', 'slices'):
            local_vals.extend(other_vals)
        else:
            #Default to putting in global slices and simplifying later
            if classes != ('global', 'slices'):
                self._change_class(key, ('global', 'slices'))
                local_vals = self.get_class_dict(('global', 'slices'))[key]
                other_vals = other._get_changed_class(key,
                                                      ('global', 'slices'),
                                                      self.slice_dim)

            #Need to interleave slices from different volumes
            n_slices = self.n_slices
            other_n_slices = other.n_slices
            shape = self.shape
            n_vols = 1
            for dim_size in shape[3:]:
                n_vols *= dim_size

            intlv = []
            loc_start = 0
            oth_start = 0
            for vol_idx in range(n_vols):
                intlv += local_vals[loc_start:loc_start + n_slices]
                intlv += other_vals[oth_start:oth_start + other_n_slices]
                loc_start += n_slices
                oth_start += other_n_slices

            self.get_class_dict(('global', 'slices'))[key] = intlv

    def _insert_non_slice(self, key, other):
        local_vals, classes = self.get_values_and_class(key)
        other_vals = other._get_changed_class(key, classes, self.slice_dim)

        if local_vals != other_vals:
            del self.get_class_dict(classes)[key]

    def _insert_sample(self, key, other, sample_base):
        local_vals, classes = self.get_values_and_class(key)
        other_vals = other._get_changed_class(key, classes, self.slice_dim)

        if classes == ('global', 'const'):
            if local_vals != other_vals:
                self._change_class(key, (sample_base, 'samples'))
                local_vals = self.get_values(key)
                other_vals = other._get_changed_class(key,
                                                      (sample_base, 'samples'),
                                                      self.slice_dim
                                                     )
                local_vals.extend(other_vals)
        elif classes == (sample_base, 'samples'):
            local_vals.extend(other_vals)
        else:
            if classes != ('global', 'slices'):
                self._change_class(key, ('global', 'slices'))
                local_vals = self.get_values(key)
                other_vals = other._get_changed_class(key,
                                                      ('global', 'slices'),
                                                      self.slice_dim)

            shape = self.shape
            n_dims = len(shape)
            if sample_base == 'time' and n_dims == 5:
                #Need to interleave values from the time points in each vector
                #component
                n_slices = self.n_slices
                slices_per_vec = n_slices * shape[3]
                oth_slc_per_vec = n_slices * other.shape[3]

                intlv = []
                loc_start = 0
                oth_start = 0
                for vec_idx in range(shape[4]):
                    intlv += local_vals[loc_start:loc_start+slices_per_vec]
                    intlv += other_vals[oth_start:oth_start+oth_slc_per_vec]
                    loc_start += slices_per_vec
                    oth_start += oth_slc_per_vec

                self.get_class_dict(('global', 'slices'))[key] = intlv
            else:
                local_vals.extend(other_vals)

#Add our extension to nibabel
nb.nifti1.extension_codes.add_codes(((dcm_meta_ecode,
                                      "dcmmeta",
                                      DcmMetaExtension),)
                                   )




SORT_GUESSES = (
    'SND.DiffusionBValue',
    'SND.EchoTime',
    'SND.InversionTime',
    'SND.RepetitionTime',
    'SND.FlipAngle',
    'SND.AcquisitionTimeStamp',
    'AcquisitionNumber',
    'InstanceNumber',
)
'''The meta data keywords used when trying to guess the sorting order.
Keys that come earlier in the list are given higher priority.'''



"""Extract meta data from a DICOM data set into human-readable and JSON compatible dict

We don't do any normalization at this level beyond converting decimal strings to float 
and making sure strings are unicode. The one exception is translating private 
mini-headers embedded in certain elements.
"""
import struct, re, warnings, enum
from collections import namedtuple
from typing import Dict, List

import pydicom
from pydicom.tag import BaseTag
from pydicom.dataset import PrivateBlock
from pydicom.datadict import keyword_for_tag, tag_for_keyword, private_dictionaries
from pydicom.charset import decode_element
from nibabel.nicom import csareader

from .dcmstack import DicomStack


# Define some common rules for ignoring DICOM elements
def ignore_pixel_data(tag, name, ds):
    return tag.group == 0x7fe0 and tag.elem in (0x8, 0x9, 0x10)


def ignore_overlay_data(tag, name, ds):
    return tag.group & 0xff00 == 0x6000 and tag.elem == 0x3000


def ignore_color_lut_data(tag, name, ds):
    return (
        tag.group == 0x28 and tag.elem in (0x1201, 0x1202, 0x1203, 0x1221, 0x1222, 0x1223)
    )


IGNORE_BINARY_RULES = (ignore_pixel_data,
                       ignore_overlay_data,
                       ignore_color_lut_data)
'''The default tuple of ignore rules for `MetaExtractor`.'''


def ignore_private(tag, name, ds):
    '''Ignore rule for `MetaExtractor` to skip all private DICOM elements'''
    return tag.group % 2 == 1


def make_ignore_unknown_private(allow_creators=None, reject_creators=None):
    """Make custom ignore rule for private elements, by default allowing "known" elements

    If the element has a name in the pydicom private dictionaries, it is considered 
    "known". Unknown elements are filtered by default, unless the "private creator" for
    that block of private elements is in `allow_creators`. Known elements will also be
    ignored if the private creator is in `reject_creators`.
    """
    allow_creators = tuple() if allow_creators is None else allow_creators
    reject_creators = tuple() if reject_creators is None else reject_creators
    if allow_creators:
        allow_re = f"({'|'.join([x for x in allow_creators])})"
    else:
        allow_re = ".^"
    if reject_creators:
        reject_re = f"({'|'.join([x for x in reject_creators])})"
    else:
        reject_re = ".^"
    def ignore_unknown_private(tag, name, ds):
        if tag.group % 2 == 0:
            return False
        toks = name.split(".")
        if re.match(allow_re, toks[0], flags=re.I):
            return False
        elif re.match(reject_re, toks[0], flags=re.I):
            return True
        return toks[1].startswith("0X")
    return ignore_unknown_private


def make_ignore_except_rule(include):
    """Make rule that ignores everying not in `include`"""
    incl_tags = set([tag_for_keyword(x) for x in include if isinstance(x, str)])
    def ignore_except(tag, name, ds):
        return tag not in incl_tags
    return ignore_except


# Define some translators for internal "mini headers" stored in private elements
Translator = namedtuple('Translator', ['name', 'tag', 'priv_creator', 'trans_func'])
'''A namedtuple for storing the four elements of a translator: a name, the
pydicom.tag.Tag that can be translated, the private creator string (optional), and
the function which takes the DICOM element and returns a dictionary.'''


def simplify_csa_dict(csa_dict):
    '''Simplify the result of nibabel.nicom.csareader.

    Parameters
    ----------
    csa_dict : dict
        The result from nibabel.nicom.csareader

    Returns
    -------
    result : dict
        Result where the keys come from the 'tags' sub dictionary of `csa_dict`.
        The values come from the 'items' within that tags sub sub dictionary.
        If items has only one element it will be unpacked from the list.
    '''
    if csa_dict is None:
        return None

    result = {}
    for tag in sorted(csa_dict['tags']):
        items = []
        for item in csa_dict['tags'][tag]['items']:
            if isinstance(item, bytes):
                try:
                    item = item.decode()
                except UnicodeDecodeError:
                    item = ""
            items.append(item)
        if len(items) == 0:
            continue
        elif len(items) == 1:
            result[tag] = items[0]
        else:
            result[tag] = items
    return result


def csa_image_trans_func(elem):
    '''Function for translating the CSA image sub header.'''
    return simplify_csa_dict(csareader.read(elem.value))


csa_image_trans = Translator('CsaImage',
                             pydicom.tag.Tag(0x29, 0x1010),
                             'SIEMENS CSA HEADER',
                             csa_image_trans_func)
'''Translator for the CSA image sub header.'''


class PhoenixParseError(Exception):
    def __init__(self, line):
        '''Exception indicating a error parsing a line from the Phoenix
        Protocol.
        '''
        self.line = line

    def __str__(self):
        return 'Unable to parse phoenix protocol line: %s' % self.line


def _parse_phoenix_line(line, str_delim='""'):
    delim_len = len(str_delim)
    #Handle most comments (not always when string literal involved)
    comment_idx = line.find('#')
    if comment_idx != -1:
        #Check if the pound sign is in a string literal
        if line[:comment_idx].count(str_delim) == 1:
            if line[comment_idx:].find(str_delim) == -1:
                raise PhoenixParseError(line)
        else:
            line = line[:comment_idx]

    #Allow empty lines
    if line.strip() == '':
        return None

    #Find the first equals sign and use that to split key/value
    equals_idx = line.find('=')
    if equals_idx == -1:
        raise PhoenixParseError(line)
    key = line[:equals_idx].strip()
    val_str = line[equals_idx + 1:].strip()

    #If there is a string literal, pull that out
    if val_str.startswith(str_delim):
        end_quote = val_str[delim_len:].find(str_delim) + delim_len
        if end_quote == -1:
            raise PhoenixParseError(line)
        elif not end_quote == len(val_str) - delim_len:
            #Make sure remainder is just comment
            if not val_str[end_quote+delim_len:].strip().startswith('#'):
                raise PhoenixParseError(line)

        return (key, val_str[2:end_quote])

    else: #Otherwise try to convert to an int or float
        val = None
        try:
            val = int(val_str)
        except ValueError:
            pass
        else:
            return (key, val)

        try:
            val = int(val_str, 16)
        except ValueError:
            pass
        else:
            return (key, val)

        try:
            val = float(val_str)
        except ValueError:
            pass
        else:
            return (key, val)

    raise PhoenixParseError(line)


def parse_phoenix_prot(prot_key, prot_val):
    '''Parse the MrPheonixProtocol string.

    Parameters
    ----------
    prot_str : str
        The 'MrPheonixProtocol' string from the CSA Series sub header.

    Returns
    -------
    prot_dict : dict
        Meta data pulled from the ASCCONV section.

    Raises
    ------
    PhoenixParseError : A line of the ASCCONV section could not be parsed.
    '''
    if prot_key == 'MrPhoenixProtocol':
        str_delim = '""'
    elif prot_key == 'MrProtocol':
        str_delim = '"'
    else:
        raise ValueError('Unknown protocol key: %s' % prot_key)
    ascconv_start = prot_val.find('### ASCCONV BEGIN ')
    ascconv_end = prot_val.find('### ASCCONV END ###')
    ascconv = prot_val[ascconv_start:ascconv_end].split('\n')[1:-1]
    result = {}
    for line in ascconv:
        parse_result = _parse_phoenix_line(line, str_delim)
        if parse_result:
            result[parse_result[0]] = parse_result[1]
    return result


def csa_series_trans_func(elem):
    '''Function for parsing the CSA series sub header.'''
    csa_dict = simplify_csa_dict(csareader.read(elem.value))

    #If there is a phoenix protocol, parse it and dump it into the csa_dict
    phx_src = None
    if 'MrPhoenixProtocol' in csa_dict:
        phx_src = 'MrPhoenixProtocol'
    elif 'MrProtocol' in csa_dict:
        phx_src = 'MrProtocol'

    if not phx_src is None:
        phoenix_dict = parse_phoenix_prot(phx_src, csa_dict[phx_src])
        del csa_dict[phx_src]
        for key, val in phoenix_dict.items():
            new_key = '%s.%s' % ('MrPhoenixProtocol', key)
            csa_dict[new_key] = val

    return csa_dict


csa_series_trans = Translator('CsaSeries',
                              pydicom.tag.Tag(0x29, 0x1020),
                              'SIEMENS CSA HEADER',
                              csa_series_trans_func)
'''Translator for parsing the CSA series sub header.'''


def xa_phoenix_trans_func(elem):
    print(elem)
    return parse_phoenix_prot("MrPhoenixProtocol", elem.value.decode())


xa_phoenix_trans = Translator(
    'MrPhoenixProtocol', pydicom.tag.Tag(0x21, 0x1119), 'SIEMENS MR SDS 01', xa_phoenix_trans_func
)


def tag_to_str(tag):
    '''Convert a DICOM tag to a string representation using the group and
    element hex values seprated by an underscore.'''
    return '%#X_%#X' % (tag.group, tag.elem)


unpack_vr_map = {'SL' : 'i',
                 'UL' : 'I',
                 'SV' : 'l',
                 'UV' : 'L',
                 'FL' : 'f',
                 'FD' : 'd',
                 'SS' : 'h',
                 'US' : 'H',
                 'US or SS' : 'H',
                 }
'''Dictionary mapping value representations to corresponding format strings for
the struct.unpack function.'''
    

class TextConverter:
    """Converter for byte strings tries to convert to text or returns None"""
    def prep_dataset(self, ds: pydicom.Dataset):
        self.encodings = pydicom.charset.convert_encodings(ds._character_set)

    def __call__(self, val):
        if isinstance(val, str):
            return val
        elif isinstance(val, bytes):
            for encoding in self.encodings:
                try:
                    return val.decode(encoding)
                except:
                    pass


_get_text = TextConverter()


class MetaExtractor(object):
    '''Callable object for extracting meta data from a dicom dataset

    Parameters
    ----------
    ignore_rules : sequence
        A sequence of callables each of which take an element tag and name, plus a 
        reference to the pydicom.Dataset, and return True if it should be ignored. If 
        None the module default is used.

    translators : sequence
        A sequence of `Translator` objects each of which can convert a
        DICOM element into a dictionary. Overrides any ignore rules. If
        None the module default is used.

    conversions : dict
        Mapping of DICOM value representation (VR) strings to callables
        that perform some conversion on the value

    warn_on_trans_except : bool
        Convert any exceptions from translators into warnings.
    '''

    IGNORE_RULES = IGNORE_BINARY_RULES + (make_ignore_unknown_private(),)

    TRANSLATORS = (csa_image_trans, csa_series_trans, xa_phoenix_trans)

    CONVERSIONS = {
        'PN' : str,
        'DS' : float,
        'IS' : int,
        'AT' : tag_to_str,
        'OW' : _get_text,
        'OB' : _get_text,
        'OW or OB' : _get_text,
        'OB or OW' : _get_text,
        'UN' : _get_text,
    }

    def __init__(
        self, 
        ignore_rules=None, 
        translators=None, 
        conversions=None,
        warn_on_trans_except=True
    ):
        if ignore_rules is None:
            self.ignore_rules = self.IGNORE_RULES
        else:
            self.ignore_rules = ignore_rules
        if translators is None:
            self.translators = self.TRANSLATORS
        else:
            self.translators = translators
        if conversions is None:
            self.conversions = self.CONVERSIONS
        else:
            self.conversions = conversions
        self.warn_on_trans_except = warn_on_trans_except

    def _get_priv_name(self, tag: BaseTag, pblocks: List[PrivateBlock]):
        '''Get the name to use for any private elements'''
        elem = creator = "Unknown"
        for pblock in pblocks:
            if pblock.block_start == tag.elem & 0xFF00:
                creator = pblock.private_creator
                priv_dict = private_dictionaries.get(creator)
                if priv_dict is None:
                    break
                key = next(iter(priv_dict))
                key = f"{key[:-2]}{tag.elem & 0xFF:02x}"
                priv_info = priv_dict.get(key)
                if priv_info is None:
                    break
                elem = priv_info[2]
                break
        if elem == "Unknown":
            elem = "%#X" % (tag.elem & 0xFF)
        else:
            elem = ''.join([t[0].upper() + t[1:] for t in elem.split()])
        return f"{creator.upper().replace(' ', '_')}.{elem}"

    def _get_elem_value(self, elem):
        '''Get the value for any non-translated elements'''
        # If the VR is implicit, we may need to unpack the values from a byte
        # string. This may require us to make an assumption about whether the
        # value is signed or not, but this is unavoidable.
        unpacked = False
        if elem.VR in unpack_vr_map and isinstance(elem.value, str):
            n_vals = len(elem.value) / struct.calcsize(unpack_vr_map[elem.VR])
            if n_vals != elem.VM:
                warnings.warn("The element's VM and the number of values do "
                              "not match.")
            if n_vals == 1:
                value = struct.unpack(unpack_vr_map[elem.VR], elem.value)[0]
            else:
                value = list(struct.unpack(unpack_vr_map[elem.VR]*n_vals,
                                           elem.value)
                            )
            unpacked = True
        else:
            value = elem.value
        n_vals = elem.VM
        # Handle any conversions
        if elem.VR in self.conversions and value is not None:
            if n_vals == 1:
                value = self.conversions[elem.VR](value)
            else:
                value = [self.conversions[elem.VR](val) for val in value]
        elif n_vals > 1 and not unpacked:
            # Make sure we don't share refences to lists in oringial dataset
            value = value[:]
        return value

    def __call__(self, dcm: pydicom.Dataset):
        '''Extract the meta data from a DICOM dataset.

        Parameters
        ----------
        dcm : pydicom.dataset.Dataset
            The DICOM dataset to extract the meta data from.

        Returns
        -------
        meta : dict
            A dictionary of extracted meta data.

        Notes
        -----
        Non-private tags use the DICOM keywords as keys. Translators have their
        name, followed by a dot, prepended to the keys of any meta elements
        they produce. Values are unchanged, except when the value
        representation is 'DS' or 'IS' (decimal/integer strings) they are
        converted to float and int types.
        '''
        # Some converters need to see full dataset before converting elements
        for conv in set(self.conversions.values()):
            if hasattr(conv, "prep_dataset"):
                conv.prep_dataset(dcm)
        # Get all the tags included in this dataset
        tags = list(dcm.keys())
        # Find private blocks in the dataset
        priv_groups = set(t.group for t in tags if t.group % 2 == 1)
        priv_blocks = {
            c : dcm.private_block(g, c) 
            for g in priv_groups
            for c in dcm.private_creators(g)
        }
        grp_blocks = {
            g : [b for b in priv_blocks.values() if b.group == g] for g in priv_groups
        }
        # Find private elements needed by translators
        # TODO: Do we want to support translating non-private elements here? Seems like
        #       just doing VR conversions is sufficient for standard meta data
        trans_map = {}
        for translator in self.translators:
            pblock = priv_blocks.get(translator.priv_creator)
            if pblock is not None:
                src_tag = pblock.get_tag(translator.tag.elem & 0xff)
                if src_tag in trans_map:
                    raise ValueError("Duplicate translators found")
                trans_map[src_tag] = translator
        result = {}
        trans_meta = {}
        charset = dcm._character_set
        for tag in tags:
            is_priv = tag.group % 2 == 1
            if is_priv:
                # Skip private creator elements
                if tag.elem & 0xFF00 == 0:
                    continue
                # If there is a translator for this element, use it
                trans = trans_map.get(tag)
                if trans is not None:
                    try:
                        meta = trans.trans_func(dcm.get(tag))
                    except Exception as e:
                        if not self.warn_on_trans_except:
                            raise
                        warnings.warn("Exception from translator %s: %s" %
                                        (trans.name,
                                        repr(str(e))))
                    else:
                        if meta:
                            trans_meta[trans.name] = meta
                    continue
                # Otherwise we construct a name for the private element
                name = self._get_priv_name(tag, grp_blocks[tag.group])
            else:
                name = keyword_for_tag(tag)
                if name == "":
                    warnings.warn("Your 'pydicom' is too old, standard tag not in datadict")
                    name = tag_to_str(tag)
            # Check if we are ignoring this element
            if any(rule(tag, name, dcm) for rule in self.ignore_rules):
                continue
            # Parse the element, skip empty values
            elem = dcm.get(tag)
            if elem.is_empty:
                continue
            # Handle elements that are sequences with recursion
            if elem.VR == "SQ":
                value = []
                for val in elem.value:
                    value.append(self(val))
                if all(x is None for x in value):
                    continue
                result[name] =  value
            # Otherwise just make sure the value is unpacked
            else:
                # Handle unicode conversion
                decode_element(elem, charset)
                value = self._get_elem_value(elem)
                if value is None:
                    continue
                result[name] =  value
        # Inject translator results
        for trans_name, meta in trans_meta.items():
            for name, value in meta.items():
                result[f"{trans_name}.{name}"] = value
        return result


# Preconfigure a few extractors with increasing levels of meta data extraction
class ExtractionLevel(enum.Enum):
    MINIMAL = "min"
    MODERATE = "mod"
    MORE = "more"
    MAX = "max"


MIN_KEYS, MIN_TRANS_NAMES = DicomStack.get_min_req_meta()
MIN_TRANSLATORS = tuple(t for t in MetaExtractor.TRANSLATORS if t.name in MIN_TRANS_NAMES)


MODERATE_KEYS = MIN_KEYS[:]
MODERATE_KEYS += [
    "Modality",
    "Manufacturer",
    "ManufacturersModelName",
    "SoftwareVersions",
    "StationName",
    "DeviceUID",
    "MagneticFieldStrength",
    "PatientID",
    "PatientsName",
    "PatientsAge",
    "PatientsSex",
    "PatientsSize",
    "PatientsWeight",
    "PatientPosition",
    "PatientSpeciesDescription",
    "PatientsBodyMassIndex",
    "BodyPartExamined",
    "StudyInstanceUID",
    "StudyID",
    "SeriesNumber",
    "StudyDescription",
    "SeriesInstanceUID",
    "ImagedNucleus",
    "ImagingFrequency",
    "SOPInstanceUID",
    "ImageType",
    "AcqusitionNumber",
    "InstanceNumber",
    "ScanningSequence",
    "SequenceVariant",
    "ScanOptions",
    "MRAcquisitionType",
    "SequenceName",
    "AngioFlag",
    "NumberOfAverages",
    "NumberOfPhaseEncodingSteps",
    "EchoTrainLength",
    "PercentSampling",
    "PercentPhaseFieldOfView",
    "PixelBandwidth",
    "TriggerTime",
    "ReceiveCoilName",
    "TransmitCoilName",
    "AcquisitionMatrix",
    "InPlanePhaseEncodingDirection",
]


EXTRACTORS = {
    ExtractionLevel.MINIMAL : MetaExtractor((make_ignore_except_rule(MIN_KEYS),), MIN_TRANSLATORS),
    ExtractionLevel.MODERATE : MetaExtractor((make_ignore_except_rule(MODERATE_KEYS),)),
    ExtractionLevel.MORE : MetaExtractor(),
    ExtractionLevel.MAX : MetaExtractor(IGNORE_BINARY_RULES),
}



"""
Command line interface for nitool.

@author: moloney
"""
from __future__ import print_function

import os, sys, argparse

from .nifti_wrapper import NiftiWrapper
import nibabel as nb

from .dcmmeta import DcmMetaExtension
from .nifti_wrapper import MissingExtensionError

prog_descrip = """Work with extended Nifti files created by dcmstack"""

def main(argv=sys.argv):
    #Setup the top level parser
    arg_parser = argparse.ArgumentParser(description=prog_descrip)
    if sys.version_info > (3, 7):
        sub_parsers = arg_parser.add_subparsers(title="Subcommands",
                                                dest='cmd',
                                                required=True)
    else:
        sub_parsers = arg_parser.add_subparsers(title="Subcommands", dest='cmd')

    #Split command
    split_help = ("Split src_nii file along a dimension. Defaults to the slice "
                  "dimension if 3D, otherwise the last dimension.")
    split_parser = sub_parsers.add_parser('split', help=split_help)
    split_parser.add_argument('src_nii', nargs=1)
    split_parser.add_argument('-d', '--dimension', default=None, type=int,
                              help=("The dimension to split along. Must be in "
                              "the range [0, 5)"))
    split_parser.add_argument('-o', '--output-format', default=None,
                              help=("Format string used to create the output "
                              "file names. Default is to prepend the index "
                              "number to the src_nii filename."))
    split_parser.set_defaults(func=split)

    #Merge Command
    merge_help = ("Merge the provided Nifti files along a dimension. Defaults "
                  "to slice, then time, and then vector.")
    merge_parser = sub_parsers.add_parser('merge', help=merge_help)
    merge_parser.add_argument('output', nargs=1)
    merge_parser.add_argument('src_niis', nargs='+')
    merge_parser.add_argument('-d', '--dimension', default=None, type=int,
                              help=("The dimension to merge along. Must be "
                              "in the range [0, 5)"))
    merge_parser.add_argument('-s', '--sort', default=None,
                              help=("Sort the source files using the provided "
                              "meta data key before merging"))
    merge_parser.add_argument('-c', '--clear-slices', action='store_true',
                              help="Clear all per slice meta data")
    merge_parser.set_defaults(func=merge)

    #Dump Command
    dump_help = "Dump the JSON meta data extension from the provided Nifti."
    dump_parser = sub_parsers.add_parser('dump', help=dump_help)
    dump_parser.add_argument('src_nii', nargs=1)
    dump_parser.add_argument('dest_json', nargs='?',
                             type=argparse.FileType('w'),
                             default=sys.stdout)
    dump_parser.add_argument('-m', '--make-empty', default=False,
                             action='store_true',
                             help="Make an empty extension if none exists")
    dump_parser.add_argument('-r', '--remove', default=False,
                             action='store_true',
                             help="Remove the extension from the Nifti file")
    dump_parser.set_defaults(func=dump)

    #Embed Command
    embed_help = "Embed a JSON extension into the Nifti file."
    embed_parser = sub_parsers.add_parser('embed', help=embed_help)
    embed_parser.add_argument('src_json', nargs='?', type=argparse.FileType('r'),
                              default=sys.stdin)
    embed_parser.add_argument('dest_nii', nargs=1)
    embed_parser.add_argument('-f', '--force-overwrite', action='store_true',
                              help="Overwrite any existing dcmmeta extension")
    embed_parser.set_defaults(func=embed)

    #Lookup command
    lookup_help = "Lookup the value for the given meta data key."
    lookup_parser = sub_parsers.add_parser('lookup', help=lookup_help)
    lookup_parser.add_argument('key', nargs=1)
    lookup_parser.add_argument('src_nii', nargs=1)
    lookup_parser.add_argument('-i', '--index',
                               help=("Use the given voxel index. The index "
                               "must be provided as a comma separated list of "
                               "integers (one for each dimension)."))
    lookup_parser.set_defaults(func=lookup)

    #Inject command
    inject_help = "Inject meta data into the JSON extension."
    inject_parser = sub_parsers.add_parser('inject', help=inject_help)
    inject_parser.add_argument('dest_nii', nargs=1)
    inject_parser.add_argument('classification', nargs=2)
    inject_parser.add_argument('key', nargs=1)
    inject_parser.add_argument('values', nargs='+')
    inject_parser.add_argument('-f', '--force-overwrite',
                               action='store_true',
                               help=("Overwrite any existing values "
                               "for the key"))
    inject_parser.add_argument('-t', '--type', default=None,
                               help="Interpret the value as this type instead "
                               "of trying to determine the type automatically")
    inject_parser.set_defaults(func=inject)

    #Parse the arguments and call the appropriate function
    args = arg_parser.parse_args(argv[1:])
    return args.func(args)

def split(args):
    src_path = args.src_nii[0]
    src_fn = os.path.basename(src_path)
    src_dir = os.path.dirname(src_path)

    src_nii = nb.load(src_path)
    try:
        src_wrp = NiftiWrapper(src_nii)
    except MissingExtensionError:
        print("No dcmmeta extension found, making empty one...")
        src_wrp = NiftiWrapper(src_nii, make_empty=True)
    for split_idx, split in enumerate(src_wrp.split(args.dimension)):
        if args.output_format:
            out_name = (args.output_format %
                        split.meta_ext.get_class_dict(('global', 'const'))
                       )
        else:
            out_name = os.path.join(src_dir, '%03d-%s' % (split_idx, src_fn))
        nb.save(split, out_name)
    return 0

def make_key_func(meta_key, index=None):
    def key_func(src_nii):
        result = src_nii.get_meta(meta_key, index)
        if result is None:
            raise ValueError('Key not found: %s' ) % meta_key
        return result

    return key_func

def merge(args):
    src_wrps = []
    for src_path in args.src_niis:
        src_nii = nb.load(src_path)
        try:
            src_wrp = NiftiWrapper(src_nii)
        except MissingExtensionError:
            print("No dcmmeta extension found, making empty one...")
            src_wrp = NiftiWrapper(src_nii, make_empty=True)
        src_wrps.append(src_wrp)

    if args.sort:
        src_wrps.sort(key=make_key_func(args.sort))

    result_wrp = NiftiWrapper.from_sequence(src_wrps, args.dimension, args.sort)

    if args.clear_slices:
        result_wrp.meta_ext.clear_slice_meta()

    out_name = (args.output[0] %
                result_wrp.meta_ext.get_class_dict(('global', 'const')))
    result_wrp.to_filename(out_name)
    return 0

def dump(args):
    src_nii = nb.load(args.src_nii[0])
    src_wrp = NiftiWrapper(src_nii, args.make_empty)
    meta_str = src_wrp.meta_ext.to_json()
    args.dest_json.write(meta_str)
    args.dest_json.write('\n')
    if args.remove:
        src_wrp.remove_extension()
        src_wrp.to_filename(args.src_nii[0])
    return 0

def check_overwrite():
    usr_input = ''
    while not usr_input in ('y', 'n'):
        usr_input = input('Existing DcmMeta extension found, overwrite? '
                              '[y/n]').lower()
    return usr_input == 'y'

def embed(args):
    ext = DcmMetaExtension.from_json(args.src_json.read())
    dest_nii = nb.load(args.dest_nii[0])
    try:
        nw = NiftiWrapper(dest_nii, False)
    except MissingExtensionError:
        hdr = dest_nii.header
    else:
        if not args.force_overwrite:
            if not check_overwrite():
                return
        nw.remove_extension()
        hdr = nw.nii_img.header
    hdr.extensions.append(ext)
    nb.save(dest_nii, args.dest_nii[0])
    return 0

def lookup(args):
    src_wrp = NiftiWrapper.from_filename(args.src_nii[0])
    index = None
    if args.index:
        index = tuple(int(idx.strip()) for idx in args.index.split(','))
    meta = src_wrp.get_meta(args.key[0], index)
    if not meta is None:
        print(meta)
    return 0

def convert_values(values, type_str=None):
    if type_str is None:
        for conv_type in (int, float):
            try:
                values = [conv_type(val) for val in values]
            except ValueError:
                pass
            else:
                break
    else:
        if type_str not in ('str', 'int', 'float'):
            raise ValueError("Unrecognized type: %s" % type_str)
        conv_type = eval(type_str)
        values = [conv_type(val) for val in values]
    if len(values) == 1:
        return values[0]
    return values

def inject(args):
    dest_nii = nb.load(args.dest_nii[0])
    dest_wrp = NiftiWrapper(dest_nii, make_empty=True)
    classification = tuple(args.classification)
    if not classification in dest_wrp.meta_ext.get_valid_classes():
        print("Invalid classification: %s" % (classification,))
        return 1
    n_vals = len(args.values)
    mult = dest_wrp.meta_ext.get_multiplicity(classification)
    if n_vals != mult:
        print(("Invalid number of values for classification. Expected "
               "%d but got %d") % (mult, n_vals))
        return 1
    key = args.key[0]
    if key in dest_wrp.meta_ext.get_keys():
        if not args.force_overwrite:
            print("Key already exists, must pass --force-overwrite")
            return 1
        else:
            curr_class = dest_wrp.meta_ext.get_classification(key)
            curr_dict = dest_wrp.meta_ext.get_class_dict(curr_class)
            del curr_dict[key]
    class_dict = dest_wrp.meta_ext.get_class_dict(classification)
    class_dict[key] = convert_values(args.values, args.type)
    nb.save(dest_nii, args.dest_nii[0])
    return 0

if __name__ == '__main__':
    sys.exit(main())



"""Command line interface to dcmstack."""
import os, sys, argparse, string
from glob import glob
from datetime import datetime

import pydicom

from . import dcmstack
from .extract import ExtractionLevel, EXTRACTORS
from .dcmstack import parse_and_group, stack_group, DicomOrdering, DEFAULT_GROUP_KEYS
from .nifti_wrapper import NiftiWrapper
from .utils import iteritems, ascii_letters, pdb_except_hook
from . import extract
from .info import __version__


prog_descrip = """Stack DICOM files from each source directory into 2D to 5D volumes, 
optionally extracting meta data.

If you use the --embed or --dump options meta data that is extracted from the DICOM 
files will be summarized and then serialized as JSON and either embedded into the ouput 
Nifti as a header extension or written separately as a JSON sidecar.
"""


prog_epilog = """IT IS YOUR RESPONSIBILITY TO KNOW IF THERE IS PRIVATE HEALTH 
INFORMATION IN THE METADATA EXTRACTED BY THIS PROGRAM."""


def parse_tags(opt_str):
    tag_strs = opt_str.split(',')
    tags = []
    for tag_str in tag_strs:
        tokens = tag_str.split('_')
        if len(tokens) != 2:
            raise ValueError('Invalid str format for tags')
        tags.append(pydicom.tag.Tag(int(tokens[0].strip(), 16),
                                  int(tokens[1].strip(), 16))
                   )
    return tags


def sanitize_path_comp(path_comp):
    result = []
    for char in path_comp:
        if not char in ascii_letters + string.digits + '-_.':
            result.append('_')
        else:
            result.append(char)
    return ''.join(result)


def _gen_src_batches(args):
    res = []
    for src in args.srcs:
        if not os.path.isdir(src):
            if not args.combine_srcs:
                yield src, [src]
            else:
                res.append(src)
        else:
            glob_str = os.path.join(src, '*')
            if args.file_ext:
                glob_str += args.file_ext
            if not args.combine_srcs:
                yield src, glob(glob_str)
            else:
                res += glob(glob_str)
    if res:
        yield "combined", res


def main(argv=sys.argv):
    #Handle command line options
    arg_parser = argparse.ArgumentParser(description=prog_descrip,
                                         epilog=prog_epilog)
    arg_parser.add_argument('srcs', nargs='*', help=('The source files / dirs'))

    input_opt = arg_parser.add_argument_group('Input options')
    input_opt.add_argument('-c', '--combine-srcs', action='store_true', 
                           help=("Combine all 'srcs' instead of processing individually"))
    input_opt.add_argument('--force-read', action='store_true', default=False,
                           help=('Try reading all files as DICOM, even if they '
                           'are missing the preamble.'))
    input_opt.add_argument('--file-ext', default='.dcm', help=('Only try reading '
                           'files with the given extension. Default: '
                           '%(default)s'))

    output_opt = arg_parser.add_argument_group('Output options')
    output_opt.add_argument('--dest-dir', default=None,
                            help=('Destination directory, defaults to the '
                            'source directory.'))
    output_opt.add_argument('-o', '--output-name', default=None,
                            help=('Python format string determining the output '
                            'filenames based on DICOM tags.'))
    output_opt.add_argument('--output-ext', default='.nii.gz',
                            help=('The extension for the output file type. '
                            'Default: %(default)s'))
    output_opt.add_argument('-d', '--dump-meta', default=False,
                            action='store_true', help=('Dump the extracted '
                            'meta data into a JSON file with the same base '
                            'name as the generated Nifti'))
    output_opt.add_argument('--embed-meta', default=False, action='store_true',
                            help=('Embed the extracted meta data into a Nifti '
                            'header extension (in JSON format).'))

    stack_opt = arg_parser.add_argument_group('Stacking Options')
    stack_opt.add_argument('-g', '--group-by', default=None,
                           help=("Comma separated list of meta data keys to "
                           "group input files into stacks with."))
    stack_opt.add_argument('--voxel-order', default='LAS',
                           help=('Order the voxels so the spatial indices '
                           'start from these directions in patient space. '
                           'The directions in patient space should be given '
                           'as a three character code: (l)eft, (r)ight, '
                           '(a)nterior, (p)osterior, (s)uperior, (i)nferior. '
                           'Passing an empty string will disable '
                           'reorientation. '
                           'Default: %(default)s'))
    stack_opt.add_argument('-t', '--time-var', default=None,
                           help=('The DICOM element keyword to use for '
                           'ordering the stack along the time dimension.'))
    stack_opt.add_argument('--vector-var', default=None,
                           help=('The DICOM element keyword to use for '
                           'ordering the stack along the vector dimension.'))
    stack_opt.add_argument('--time-order', default=None,
                           help=('Provide a text file with the desired order '
                           'for the values (one per line) of the attribute '
                           'used as the time variable. This option is rarely '
                           'needed.'))
    stack_opt.add_argument('--vector-order', default=None,
                           help=('Provide a text file with the desired order '
                           'for the values (one per line) of the attribute '
                           'used as the vector variable. This option is rarely '
                           'needed.'))

    extr_opt = arg_parser.add_argument_group('Meta Data Extraction Options')
    extr_opt.add_argument("-l", "--level", default="more", 
                          help=("Control the amount of meta data extracted. Higher "
                          "levels will slow down conversion. Options: "
                          f"{'/'.join(e.value for e in ExtractionLevel)}"))
    extr_opt.add_argument('--allow-pcreator', action='append', 
                          help=("Extract all meta data in matching 'Private Creator' "
                          "blocks, even if the elem name is unknown. Can be regex."))
    extr_opt.add_argument('--reject-pcreator', action='append', 
                          help=("Skip all meta data in matching 'Private Creator' "
                          "blocks, even if the elem name is known. Can be regex."))
    extr_opt.add_argument('--list-translators', default=False,
                          action='store_true', help=('List enabled translators '
                          'and exit'))
    
    filt_opt = arg_parser.add_argument_group('Meta Data Filtering Options')
    filt_opt.add_argument('-i', '--include-regex', action='append',
                          help=('Include any meta data where the key matches '
                          'the provided regular expression. This will override '
                          'any exclude expressions. Applies to all meta data.'))
    filt_opt.add_argument('-e', '--exclude-regex', action='append',
                          help=('Exclude any meta data where the key matches '
                          'the provided regular expression. This will '
                          'supplement the default exclude expressions. Applies '
                          'to all meta data.'))
    filt_opt.add_argument('--default-regexes', default=False,
                          action='store_true',
                          help=('Print the list of default include and exclude '
                          'regular expressions and exit.'))

    gen_opt = arg_parser.add_argument_group('General Options')
    gen_opt.add_argument('-v', '--verbose',  default=False, action='store_true',
                         help=('Print additional information.'))
    gen_opt.add_argument('--strict', default=False, action='store_true',
                         help=('Fail on the first exception instead of '
                         'showing a warning.'))
    gen_opt.add_argument('--pdb', default=False, action='store_true',
                         help=('Enter debugger on unhandled exceptions.'))
    gen_opt.add_argument('--version', default=False, action='store_true',
                         help=('Show the version and exit.'))
    
    args = arg_parser.parse_args(argv[1:])
    if args.pdb:
        if sys.stderr.isatty():
            sys.excepthook = pdb_except_hook
        else:
            print("Ignoring '--pdb' in non-interactive context")
    if args.version:
        print(__version__)
        return 0
    #Check if we are just listing the translators
    if args.list_translators:
        for translator in extract.default_translators:
            print('%s -> %s' % (translator.tag, translator.name))
        return 0
    #Check if we are just listing the default exclude regular expressions
    if args.default_regexes:
        print('Default exclude regular expressions:')
        for regex in dcmstack.default_key_excl_res:
            print('\t' + regex)
        print('Default include regular expressions:')
        for regex in dcmstack.default_key_incl_res:
            print('\t' + regex)
        return 0
    if len(args.srcs) == 0:
        arg_parser.error('No sources were provided')
    # If we are generating meta data setup the extractors, otherwise use minimal one
    gen_meta = args.embed_meta or args.dump_meta
    if gen_meta:
        extractor = extract.EXTRACTORS[extract.ExtractionLevel(args.level.lower())]
        if args.allow_pcreator or args.reject_pcreator:
            kwargs = {}
            if args.allow_pcreator:
                kwargs["allow_creators"] = args.allow_pcreator
            if args.reject_pcreator:
                kwargs["reject_creators"] = args.reject_pcreator
            priv_rule = extract.make_ignore_unknown_private(**kwargs)
            extractor.ignore_rules = extract.IGNORE_BINARY_RULES + (priv_rule,)
    else:
        extractor = extract.EXTRACTORS[ExtractionLevel.MINIMAL]
    # Add include/exclude regexes to meta filter
    include_regexes = dcmstack.default_key_incl_res
    if args.include_regex:
        include_regexes += args.include_regex
    exclude_regexes = dcmstack.default_key_excl_res
    if args.exclude_regex:
        exclude_regexes += args.exclude_regex
    meta_filter = dcmstack.make_key_regex_filter(exclude_regexes, include_regexes)
    # Figure out time and vector ordering
    if args.time_var:
        if args.time_order:
            order_file = open(args.time_order)
            abs_order = [line.strip() for line in order_file.readlines()]
            order_file.close()
            time_order = DicomOrdering(args.time_var, abs_order, True)
        else:
            time_order = DicomOrdering(args.time_var)
    else:
        time_order = None
    if args.vector_var:
        if args.vector_order:
            order_file = open(args.vector_order)
            abs_order = [line.strip() for line in order_file.readlines()]
            order_file.close()
            vector_order = DicomOrdering(args.vector_var, abs_order, True)
        else:
            vector_order = DicomOrdering(args.vector_var)
    else:
        vector_order = None
    #Handle group-by option
    if not args.group_by is None:
        group_by = args.group_by.split(',')
    else:
        group_by = DEFAULT_GROUP_KEYS

    # Process files in batches
    for src, src_paths in _gen_src_batches(args):
        if args.verbose:
            print(f"Processing source ({len(src_paths)} files): {src}")
            start = datetime.now()
            src_start = start
        # Group the files in this batch
        groups = parse_and_group(src_paths,
                                 group_by,
                                 extractor,
                                 args.force_read,
                                 not args.strict,
                                )
        if args.verbose:
            delta = datetime.now() - start
            print(f"Parsed input files and found {len(groups)} group(s) of DICOM images (took {delta})")
        if len(groups) == 0:
            print("No DICOM files found in %s" % src)
        out_idx = 0
        generated_outs = set()
        for key, group in iteritems(groups):
            if args.verbose:
                start = datetime.now()
            stack = stack_group(group,
                                warn_on_except=not args.strict,
                                time_order=time_order,
                                vector_order=vector_order,
                                meta_filter=meta_filter)
            if args.verbose:
                delta = datetime.now() - start
                print(f"DicomStack creation took: {delta}")
                start = datetime.now()
            nii = stack.to_nifti(args.voxel_order, gen_meta)
            if args.verbose:
                delta = datetime.now() - start
                print(f"Nifti creation took: {delta}")
            meta = stack._files_info[0][0].meta_ext.get_class_dict(("global", "const"))
            #Build an appropriate output format string if none was specified
            if args.output_name is None:
                out_fmt = []
                if 'SeriesNumber' in meta:
                    out_fmt.append('%(SeriesNumber)03d')
                if 'SND.AcquisitionDescription' in meta:
                    out_fmt.append('%(SND.AcquisitionDescription)s')
                else:
                    out_fmt.append('series')
                out_fmt = '-'.join(out_fmt)
            else:
                out_fmt = args.output_name
            # Get the output filename from the format string, make sure the result is 
            # unique for this invocation
            out_fn = sanitize_path_comp(out_fmt % meta)
            if out_fn in generated_outs:
                out_fn += '-%03d' % out_idx
            generated_outs.add(out_fn)
            out_idx += 1
            out_fn = out_fn + args.output_ext
            if args.dest_dir:
                out_dir =args.dest_dir
            else:
                if src == 'combined':
                    out_dir = args.srcs[0]
                else:
                    out_dir = src
                if not out_dir.is_dir():
                    out_dir = out_dir.parent
            out_path = os.path.join(out_dir, out_fn)
            if args.dump_meta:
                nii_wrp = NiftiWrapper(nii)
                path_tokens = out_path.split('.')
                if path_tokens[-1] == 'gz':
                    path_tokens = path_tokens[:-1]
                if path_tokens[-1] == 'nii':
                    path_tokens = path_tokens[:-1]
                meta_path = '.'.join(path_tokens + ['json'])
                if args.verbose:
                    print(f"Dumping meta data to external JSON: {meta_path}")
                out_file = open(meta_path, 'w')
                out_file.write(nii_wrp.meta_ext.to_json())
                out_file.close()
                if not args.embed_meta:
                    nii_wrp.remove_extension()
                del nii_wrp
            if args.verbose:
                print("Writing out stack to path %s" % out_path)
                start = datetime.now()
            nii.to_filename(out_path)
            if args.verbose:
                end = datetime.now()
                print(f"Writing / compression took: {end - start}")
                print(F"Total conversion for source took: {end - src_start}")
            del key
            del group
            del stack
            del meta
            del nii
        del groups

    return 0

if __name__ == '__main__':
    sys.exit(main())



"""
Package for stacking DICOM images into multi dimensional volumes, extracting
the DICOM meta data, converting the result to Nifti files with the meta
data stored in a header extension, and work with these extended Nifti files.
"""
from .info import __version__
from .dcmstack import *



"""Provide wrapper for a nibabel Nifti image with a meta data extension

Also provides conversion from single DICOM to equivalent Nifti plus meta data
"""
import itertools
import warnings
from typing import List, Sequence, Optional, Union

import numpy as np
import nibabel as nb
from nibabel.spatialimages import HeaderDataError
from nibabel.nifti1 import Nifti1Image, Nifti1Extension
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    from nibabel.nicom.dicomwrappers import wrapper_from_data

from .dcmmeta import DcmMetaExtension, InvalidExtensionError, dcm_meta_ecode


class MissingExtensionError(Exception):
    '''Exception denoting that there is no DcmMetaExtension in the Nifti header.
    '''
    def __str__(self):
        return 'No dcmmeta extension found.'


def patch_dcm_ds_is(dcm):
    '''Convert all elements in `dcm` with VR of 'DS' or 'IS' to floats and ints.
    This is a hackish work around for the backwards incompatibility of pydicom
    0.9.7 and should not be needed once nibabel is updated.
    '''
    for elem in dcm:
        if elem.VM == 1:
            if elem.VR in ('DS', 'IS'):
                if elem.value == '':
                    continue
                if elem.VR == 'DS':
                    elem.VR = 'FD'
                    elem.value = float(elem.value)
                else:
                    elem.VR = 'SL'
                    elem.value = int(elem.value)
        else:
            if elem.VR in ('DS', 'IS'):
                if elem.value == '':
                    continue
                if elem.VR == 'DS':
                    elem.VR = 'FD'
                    elem.value = [float(val) for val in elem.value]
                else:
                    elem.VR = 'SL'
                    elem.value = [int(val) for val in elem.value]


def gen_simplified_sequences(meta_dict):
    """Get rid of useless nesting of meta data from multiframe DICOM"""
    for k, v in meta_dict.items():
        if isinstance(v, list):
            if len(v) == 0:
                continue
            if len(v) == 1:
                for sub_key, sub_val in v[0].items():
                    yield sub_key, sub_val
                continue
        yield k, v
    

def _get_fg_const_and_varying(fg_seqs):
    varying = {}
    for idx, fg_seq in enumerate(fg_seqs):
        for k, v in gen_simplified_sequences(fg_seq):
            if k not in varying:
                if idx == 0:
                    varying[k] = [v]
                else:
                    varying[k] = [None] * idx
                    varying[k].append(v)
            else:
                n_vals = len(varying[k])
                if n_vals != idx:
                    varying[k] += [None] * (idx - n_vals)
                varying[k].append(v)
    const = {}
    for k, vals in varying.items():
        if len(vals) != len(fg_seqs):
            vals += [None] * (len(fg_seqs) - len(vals))
        if all(x == vals[0] for x in vals):
            const[k] = vals[0]
    for k in const:
        del varying[k]
    return const, varying

class NiftiWrapper(object):
    '''Wraps a Nifti1Image object containing a DcmMeta header extension.
    Provides access to the meta data and the ability to split or merge the
    data array while updating the meta data.

    Parameters
    ----------
    nii_img : nibabel.nifti1.Nifti1Image
        The Nifti1Image to wrap.

    make_empty : bool
        If True an empty DcmMetaExtension will be created if none is found.

    Raises
    ------
    MissingExtensionError
        No valid DcmMetaExtension was found.

    ValueError
        More than one valid DcmMetaExtension was found.
    '''

    def __init__(self, nii_img: Nifti1Image, make_empty: bool = False):
        self.nii_img = nii_img
        hdr = nii_img.header
        ext = None
        self.meta_ext: Nifti1Extension = None
        for extension in hdr.extensions:
            if extension.get_code() == dcm_meta_ecode:
                try:
                    extension.check_valid()
                except InvalidExtensionError as e:
                    print("Found candidate extension, but invalid: %s" % e)
                else:
                    if not ext is None:
                        raise ValueError('More than one valid DcmMeta '
                                         'extension found.')
                    ext = extension
        if ext is None:
            if make_empty:
                slice_dim = hdr.get_dim_info()[2]
                ext = DcmMetaExtension.make_empty(
                    self.nii_img.shape,
                    hdr.get_best_affine(),
                    None,
                    slice_dim
                )
                hdr.extensions.append(ext)
            else:
                raise MissingExtensionError
        self.meta_ext: Nifti1Extension = ext

    def __getitem__(self, key):
        '''Get the value for the given meta data key. Only considers meta data
        that is globally constant. To access varying meta data you must use the
        method 'get_meta'.'''
        return self.meta_ext.get_class_dict(('global', 'const'))[key]

    def meta_valid(self, classification):
        '''Return true if the meta data with the given classification appears
        to be valid for the wrapped Nifti image. Considers the shape and
        orientation of the image and the meta data extension.'''
        if classification == ('global', 'const'):
            return True

        img_shape = self.nii_img.shape
        meta_shape = self.meta_ext.shape
        if classification == ('vector', 'samples'):
            return meta_shape[4:] == img_shape[4:]
        if classification == ('time', 'samples'):
            return meta_shape[3:] == img_shape[3:]

        hdr = self.nii_img.header
        if self.meta_ext.n_slices != hdr.get_n_slices():
            return False

        slice_dim = hdr.get_dim_info()[2]
        slice_dir = self.nii_img.affine[slice_dim, :3]
        slices_aligned = np.allclose(slice_dir,
                                     self.meta_ext.slice_normal,
                                     atol=1e-6)

        if classification == ('time', 'slices'):
            return slices_aligned
        if classification == ('vector', 'slices'):
            return meta_shape[3] == img_shape[3] and slices_aligned
        if classification == ('global', 'slices'):
            return meta_shape[3:] == img_shape[3:] and slices_aligned

    def get_meta(self, key, index=None, default=None):
        '''Return the meta data value for the provided `key`.

        Parameters
        ----------
        key : str
            The meta data key.

        index : tuple
            The voxel index we are interested in.

        default
            This will be returned if the meta data for `key` is not found.

        Returns
        -------
        value
            The meta data value for the given `key` (and optionally `index`)

        Notes
        -----
        The per-sample and per-slice meta data will only be considered if the
        `samples_valid` and `slices_valid` methods return True (respectively),
        and an `index` is specified.
        '''
        #Get the value(s) and classification for the key
        values, classes = self.meta_ext.get_values_and_class(key)
        if classes is None:
            return default

        #Check if the value is constant
        if classes == ('global', 'const'):
            return values

        #Check if the classification is valid
        if not self.meta_valid(classes):
            return default

        #If an index is provided check the varying values
        if not index is None:
            #Test if the index is valid
            shape = self.nii_img.shape
            if len(index) != len(shape):
                raise IndexError('Incorrect number of indices.')
            for dim, ind_val in enumerate(index):
                if not 0 <= ind_val < shape[dim]:
                    raise IndexError('Index is out of bounds.')

            #First try per time/vector sample values
            if classes == ('time', 'samples'):
                return values[index[3]]
            if classes == ('vector', 'samples'):
                return values[index[4]]

            #Finally, if aligned, try per-slice values
            slice_dim = self.nii_img.header.get_dim_info()[2]
            n_slices = shape[slice_dim]
            if classes == ('global', 'slices'):
                val_idx = index[slice_dim]
                for count, idx_val in enumerate(index[3:]):
                    val_idx += idx_val * n_slices
                    n_slices *= shape[count+3]
                return values[val_idx]
            elif classes == ('time', 'slices'):
                val_idx = index[slice_dim]
                return values[val_idx]
            elif classes == ('vector', 'slices'):
                val_idx = index[slice_dim]
                val_idx += index[3]*n_slices
                return values[val_idx]

        return default

    def remove_extension(self):
        '''Remove the DcmMetaExtension from the header of nii_img. The
        attribute `meta_ext` will still point to the extension.'''
        hdr = self.nii_img.header
        target_idx = None
        for idx, ext in enumerate(hdr.extensions):
            if id(ext) == id(self.meta_ext):
                target_idx = idx
                break
        else:
            raise IndexError('Extension not found in header')
        del hdr.extensions[target_idx]
        # Nifti1Image.update_header will increase this if necessary
        hdr['vox_offset'] = 0

    def replace_extension(self, dcmmeta_ext: Nifti1Extension):
        '''Replace the DcmMetaExtension.

        Parameters
        ----------
        dcmmeta_ext : DcmMetaExtension
            The new DcmMetaExtension.

        '''
        self.remove_extension()
        self.nii_img.header.extensions.append(dcmmeta_ext)
        self.meta_ext = dcmmeta_ext

    def split(self, dim=None):
        '''Generate splits of the array and meta data along the specified
        dimension.

        Parameters
        ----------
        dim : int
            The dimension to split the voxel array along. If None it will
            prefer the vector, then time, then slice dimensions.

        Returns
        -------
        result
            Generator which yields a NiftiWrapper result for each index
            along `dim`.

        '''
        shape = self.nii_img.shape
        data = np.asanyarray(self.nii_img.dataobj)
        header = self.nii_img.header
        slice_dim = header.get_dim_info()[2]

        #If dim is None, choose the vector/time/slice dim in that order
        if dim is None:
            dim = len(shape) - 1
            if dim == 2:
                if slice_dim is None:
                    raise ValueError("Slice dimension is not known")
                dim = slice_dim

        #If we are splitting on a spatial dimension, we need to update the
        #translation
        trans_update = None
        if dim < 3:
            trans_update = header.get_best_affine()[:3, dim]

        split_hdr = header.copy()
        slices = [slice(None)] * len(shape)
        for idx in range(shape[dim]):
            #Grab the split data, get rid of trailing singular dimensions
            if dim >= 3 and dim == len(shape) - 1:
                slices[dim] = idx
            else:
                slices[dim] = slice(idx, idx+1)

            split_data = data[tuple(slices)].copy()

            #Update the translation in any affines if needed
            if not trans_update is None and idx != 0:
                qform = split_hdr.get_qform()
                if not qform is None:
                    qform[:3, 3] += trans_update
                    split_hdr.set_qform(qform)
                sform = split_hdr.get_sform()
                if not sform is None:
                    sform[:3, 3] += trans_update
                    split_hdr.set_sform(sform)

            #Create the initial Nifti1Image object
            split_nii = Nifti1Image(
                split_data,
                split_hdr.get_best_affine(),
                header=split_hdr
            )

            #Replace the meta data with the appropriate subset
            meta_dim = dim
            if dim == slice_dim:
                meta_dim = self.meta_ext.slice_dim
            split_meta = self.meta_ext.get_subset(meta_dim, idx)
            result = NiftiWrapper(split_nii)
            result.replace_extension(split_meta)

            yield result

    def to_filename(self, out_path):
        '''Write out the wrapped Nifti to a file

        Parameters
        ----------
        out_path : str
            The path to write out the file to

        Notes
        -----
        Will check that the DcmMetaExtension is valid before writing the file.
        '''
        self.meta_ext.check_valid()
        self.nii_img.to_filename(out_path)

    @classmethod
    def from_filename(klass, path):
        '''Create a NiftiWrapper from a file.

        Parameters
        ----------
        path : str
            The path to the Nifti file to load.
        '''
        return klass(nb.load(path))

    @classmethod
    def from_dicom_wrapper(klass, dcm_wrp, meta_dict=None):
        '''Create a NiftiWrapper from a nibabel DicomWrapper.

        Parameters
        ----------
        dcm_wrap : nicom.dicomwrappers.DicomWrapper
            The dataset to convert into a NiftiWrapper.

        meta_dict : dict
            An optional dictionary of meta data extracted from `dcm_data`. See
            the `extract` module for generating this dict.
        '''
        shape = dcm_wrp.image_shape
        n_dims = len(shape)
        if n_dims > 4:
            raise ValueError("5D+ multiframe not supported")
        # Figure out any data rescaling
        scale_factors = dcm_wrp.scale_factors
        if dcm_wrp.is_multiframe and scale_factors.shape[1] > 1:
            slope, inter = 1, 0
            data = dcm_wrp.get_data()
        else:
            slope, inter = scale_factors[0, :]
            data = dcm_wrp.get_unscaled_data()
        # The Nifti patient space flips the x and y directions
        affine = np.dot(np.diag([-1., -1., 1., 1.]), dcm_wrp.affine)
        n_vols = 1
        if n_dims == 2:
            data = data.reshape(data.shape + (1,))
            slices_per_vol = 1
        elif n_dims >= 3:
            data = np.squeeze(data)
            slices_per_vol = shape[2]
        if n_dims == 4:
            n_vols = shape[3]
        # Create the nifti image and set header data
        nii_img = Nifti1Image(data, affine)
        hdr = nii_img.header
        if (slope, inter) != (1.0, 0.0):
            hdr.set_slope_inter(slope, inter)
        hdr.set_xyzt_units('mm', 'sec')
        # Determine phase encoding direction, set dimension info in header
        phase_info = None
        if dcm_wrp.is_multiframe:
            phase_info = dcm_wrp.shared.get("MRFOVGeometrySequence")
            if phase_info is None and "MRFOVGeometrySequence" in dcm_wrp.frames[0]:
                phase_info = [f.get("MRFOVGeometrySequence")[0] for f in dcm_wrp.frames]
        if phase_info is None:
            phase_info = [dcm_wrp]
        phase_dirs = set(d.get('InPlanePhaseEncodingDirection') for d in phase_info)
        if len(phase_dirs) > 1:
            phase_dir = None
        else:
            phase_dir = phase_dirs.pop()
        dim_info = {'freq' : None, 'phase' : None, 'slice' : 2}
        if phase_dir:
            if phase_dir == 'ROW':
                dim_info['phase'], dim_info['freq'] = 1, 0
            else:
                dim_info['phase'], dim_info['freq'] = 0, 1
        hdr.set_dim_info(**dim_info)
        # Create result and embed any provided meta data
        result = klass(nii_img, make_empty=True)
        result.meta_ext.reorient_transform = np.eye(4)
        if meta_dict:
            if not dcm_wrp.is_multiframe:
                global_meta = result.meta_ext.get_class_dict(('global', 'const'))
                global_meta.update(meta_dict)
                if dcm_wrp.is_mosaic:
                    # For mosaic images we move a few elems that provide per-slice meta
                    slice_meta = result.meta_ext.get_class_dict(('global', 'slices'))
                    for key in (
                        "SIEMENS_MR_HEADER.MosaicRefAcqTimes",
                        "CsaImage.MosaicRefAcqTimes",
                        "SourceImageSequence",
                    ):
                        vals = global_meta.get(key)
                        if vals is not None:
                            slice_meta[key] = vals
                            del global_meta[key]
            else:
                # Unpack and sort meta data from multiframe file that varies
                global_meta = meta_dict.copy()
                del global_meta["SharedFunctionalGroupsSequence"]
                del global_meta["PerFrameFunctionalGroupsSequence"]
                assert len(meta_dict["SharedFunctionalGroupsSequence"]) == 1
                for k, v in gen_simplified_sequences(
                    meta_dict["SharedFunctionalGroupsSequence"][0]
                ):
                    if k in global_meta:
                        if global_meta[k] == v:
                            continue
                        k = f"Shared.{k}"
                    global_meta[k] = v
                fg_seqs = meta_dict.get("PerFrameFunctionalGroupsSequence")
                sorted_indices = dcm_wrp.frame_order
                fg_keys = set()
                slice_keys = set()
                vol_results = []
                for vol_idx in range(n_vols):
                    start = vol_idx * slices_per_vol
                    end = start + slices_per_vol
                    vol_res = _get_fg_const_and_varying(
                        [fg_seqs[x] for x in sorted_indices[start:end]]
                    )
                    vol_results.append(vol_res)
                    for res in vol_res:
                        for k in res:
                            fg_keys.add(k)
                    for k in vol_res[1]:
                        slice_keys.add(k)
                global_slices = result.meta_ext.get_class_dict(('global', 'slices'))
                if n_vols > 1:
                    for vconst, vslices in vol_results:
                        for k in fg_keys:
                            if k not in vconst and k not in vslices:
                                vconst[k] = None
                        for k in slice_keys:
                            if k in vconst:
                                vslices[k] = [vconst[k]] * slices_per_vol
                                del vconst[k]
                    time_samples = result.meta_ext.get_class_dict(('time', 'samples'))
                    time_slices = result.meta_ext.get_class_dict(('time', 'slices'))
                    for k, first_val in vol_results[0][0].items():
                        dest_key = k if k not in global_meta else f"PerFrame.{k}"
                        vals = [vconst[k] for vconst, _ in vol_results]
                        if all(v == first_val for v in vals):
                            if k in global_meta and global_meta[k] == first_val:
                                continue
                            global_meta[dest_key] = first_val
                        else:
                            time_samples[dest_key] = vals
                    for k, first_val in vol_results[0][1].items():
                        dest_key = k if k not in global_meta else f"PerFrame.{k}"
                        vals = [vslices[k] for _, vslices in vol_results]
                        if all(v == first_val for v in vals):
                            time_slices[dest_key] = first_val
                        else:
                            global_slices[dest_key] = list(itertools.chain(*vals))
                else:
                    vconst, vslices = vol_results[0]
                    for k, val in vconst.items():
                        if k in global_meta:
                            if global_meta[k] != val:
                                global_meta[f"PerFrame.{k}"] = val
                        else:
                            global_meta[k] = val
                    for k, vals in vslices.items():
                        if k in global_meta:
                            global_slices[f"PerFrame.{k}"] = vals
                        else:
                            global_slices[k] = vals
                result.meta_ext.get_class_dict(('global', 'const')).update(global_meta)
        return result

    @classmethod
    def from_dicom(klass, dcm_data, meta_dict=None):
        '''Create a NiftiWrapper from a single DICOM dataset.

        Parameters
        ----------
        dcm_data : dicom.dataset.Dataset
            The DICOM dataset to convert into a NiftiWrapper.

        meta_dict : dict
            An optional dictionary of meta data extracted from `dcm_data`. See
            the `extract` module for generating this dict.

        '''
        dcm_wrp = wrapper_from_data(dcm_data)
        return klass.from_dicom_wrapper(dcm_wrp, meta_dict)

    @classmethod
    def from_sequence(
        klass,
        seq: Sequence["NiftiWrapper"],
        dim: Optional[int] = None,
        dim_key: Optional[str] = None,
    ) -> "NiftiWrapper":
        '''Create a NiftiWrapper by joining a sequence of NiftiWrapper objects
        along the given dimension.

        Parameters
        ----------
        seq : sequence
            The sequence of NiftiWrapper objects.

        dim : int
            The dimension to join the NiftiWrapper objects along. If None,
            2D inputs will become 3D, 3D inputs will become 4D, and 4D inputs
            will become 5D.

        Returns
        -------
        result : NiftiWrapper
            The merged NiftiWrapper with updated meta data.
        '''
        n_inputs = len(seq)
        first_input = seq[0]
        first_nii = first_input.nii_img
        first_hdr = first_nii.header
        shape = first_nii.shape
        affine = first_nii.affine.copy()
        #If dim is None, choose a sane default
        if dim is None:
            if len(shape) == 3:
                singular_dim = None
                for dim_idx, dim_size in enumerate(shape):
                    if dim_size == 1:
                        singular_dim = dim_idx
                if singular_dim is None:
                    dim = 3
                else:
                    dim = singular_dim
            if len(shape) == 4:
                dim = 4
        else:
            if not 0 <= dim < 5:
                raise ValueError("The argument 'dim' must be in the range "
                                 "[0, 5).")
            if dim < len(shape) and shape[dim] != 1:
                raise ValueError('The dimension must be singular or not exist')
        assert dim is not None
        #Pull out the three axes vectors for validation of other input affines
        axes = []
        for axis_idx in range(3):
            axis_vec = affine[:3, axis_idx]
            if axis_idx == dim:
                axis_vec = axis_vec.copy()
                axis_vec /= np.sqrt(np.dot(axis_vec, axis_vec))
            axes.append(axis_vec)
        #Pull out the translation
        trans = affine[:3, 3]
        #Determine the shape / dtype of the result data array and create it
        result_shape = list(shape)
        while dim >= len(result_shape):
            result_shape.append(1)
        result_shape[dim] = n_inputs
        result_dtype = max(input_wrp.nii_img.get_data_dtype()
                           for input_wrp in seq)
        result_data = np.empty(result_shape, dtype=result_dtype)
        #Start with the header info from the first input
        hdr_info = {'qform' : first_hdr.get_qform(),
                    'qform_code' : first_hdr['qform_code'],
                    'sform' : first_hdr.get_sform(),
                    'sform_code' : first_hdr['sform_code'],
                    'dim_info' : list(first_hdr.get_dim_info()),
                    'xyzt_units' : list(first_hdr.get_xyzt_units()),
                   }
        try:
            hdr_info['slice_duration'] = first_hdr.get_slice_duration()
        except HeaderDataError:
            hdr_info['slice_duration'] = None
        try:
            hdr_info['intent'] = first_hdr.get_intent()
        except HeaderDataError:
            hdr_info['intent'] = None
        try:
            hdr_info['slice_times'] = first_hdr.get_slice_times()
        except HeaderDataError:
            hdr_info['slice_times'] = None
        #Fill the data array, check header consistency
        data_slices: List[Union[slice, int]] = [slice(None)] * len(result_shape)
        for dim_idx, dim_size in enumerate(result_shape):
            if dim_size == 1:
                data_slices[dim_idx] = 0
        last_trans = None #Keep track of the translation from last input
        for input_idx in range(n_inputs):
            input_wrp = seq[input_idx]
            input_nii = input_wrp.nii_img
            input_aff = input_nii.affine
            input_hdr = input_nii.header
            #Check that the affines match appropriately
            for axis_idx, axis_vec in enumerate(axes):
                in_vec = input_aff[:3, axis_idx]
                #If we are joining on this dimension
                if axis_idx == dim:
                    #Allow scaling difference as it will be updated later
                    in_vec = in_vec.copy()
                    in_vec /= np.sqrt(np.dot(in_vec, in_vec))
                    in_trans = input_aff[:3, 3]
                    if not last_trans is None:
                        #Must be translated along the axis
                        trans_diff = in_trans - last_trans
                        if not np.allclose(trans_diff, 0.0):
                            trans_diff /= np.sqrt(np.dot(trans_diff, trans_diff))
                        if (np.allclose(trans_diff, 0.0) or
                            not np.allclose(np.dot(trans_diff, in_vec),
                                            1.0,
                                            atol=1e-6)
                           ):
                            raise ValueError("Slices must be translated along the "
                                             "normal direction")
                    #Update reference to last translation
                    last_trans = in_trans
                #Check that axis vectors match
                if not np.allclose(in_vec, axis_vec, atol=5e-4):
                    raise ValueError("Cannot join images with different "
                                     "orientations.")
            data_slices[dim] = input_idx
            result_data[tuple(data_slices)] = np.asanyarray(input_nii.dataobj).squeeze()
            if input_idx != 0:
                if (hdr_info['qform'] is None or
                    input_hdr.get_qform() is None or
                    not np.allclose(input_hdr.get_qform(), hdr_info['qform'])
                   ):
                    hdr_info['qform'] = None
                if input_hdr['qform_code'] != hdr_info['qform_code']:
                    hdr_info['qform_code'] = None
                if (hdr_info['sform'] is None or
                    input_hdr.get_sform() is None or
                    not np.allclose(input_hdr.get_sform(), hdr_info['sform'])
                   ):
                    hdr_info['sform'] = None
                if input_hdr['sform_code'] != hdr_info['sform_code']:
                    hdr_info['sform_code'] = None
                in_dim_info = list(input_hdr.get_dim_info())
                if in_dim_info != hdr_info['dim_info']:
                    for idx in range(3):
                        if in_dim_info[idx] != hdr_info['dim_info'][idx]:
                            hdr_info['dim_info'][idx] = None
                in_xyzt_units = list(input_hdr.get_xyzt_units())
                if in_xyzt_units != hdr_info['xyzt_units']:
                    for idx in range(2):
                        if in_xyzt_units[idx] != hdr_info['xyzt_units'][idx]:
                            hdr_info['xyzt_units'][idx] = None
                try:
                    if input_hdr.get_slice_duration() != hdr_info['slice_duration']:
                        hdr_info['slice_duration'] = None
                except HeaderDataError:
                    hdr_info['slice_duration'] = None
                try:
                    if input_hdr.get_intent() != hdr_info['intent']:
                        hdr_info['intent'] = None
                except HeaderDataError:
                    hdr_info['intent'] = None
                try:
                    if input_hdr.get_slice_times() != hdr_info['slice_times']:
                        hdr_info['slice_times'] = None
                except HeaderDataError:
                    hdr_info['slice_times'] = None
        #If we joined along a spatial dim, rescale the appropriate axis
        scaled_dim_dir = None
        if dim < 3:
            scaled_dim_dir = seq[1].nii_img.affine[:3, 3] - trans
            affine[:3, dim] = scaled_dim_dir
        #Create the resulting Nifti and wrapper
        result_nii = Nifti1Image(result_data, affine)
        result_hdr = result_nii.header
        #Update the header with any info that is consistent across inputs
        if hdr_info['qform'] is not None and hdr_info['qform_code'] is not None:
            if not scaled_dim_dir is None:
                hdr_info['qform'][:3, dim] = scaled_dim_dir
            result_nii.set_qform(hdr_info['qform'],
                                 int(hdr_info['qform_code']),
                                 update_affine=True)
        if hdr_info['sform'] is not None and hdr_info['sform_code'] is not None:
            if not scaled_dim_dir is None:
                hdr_info['sform'][:3, dim] = scaled_dim_dir
            result_nii.set_sform(hdr_info['sform'],
                                 int(hdr_info['sform_code']),
                                 update_affine=True)
        if hdr_info['dim_info'] is not None:
            result_hdr.set_dim_info(*hdr_info['dim_info'])
            slice_dim = hdr_info['dim_info'][2]
        else:
            slice_dim = None
        if hdr_info['intent'] is not None:
            result_hdr.set_intent(*hdr_info['intent'])
        if hdr_info['xyzt_units'] is not None:
            result_hdr.set_xyzt_units(*hdr_info['xyzt_units'])
        if hdr_info['slice_duration'] is not None:
            result_hdr.set_slice_duration(hdr_info['slice_duration'])
        if hdr_info['slice_times'] is not None:
            result_hdr.set_slice_times(hdr_info['slice_times'])
        # Create the meta data extension and insert it
        result_ext = DcmMetaExtension.from_sequence(
            [elem.meta_ext for elem in seq],
            dim,
            affine,
            slice_dim,
        )
        if dim_key is not None:
            if dim < 3:
                warnings.warn("Ignoring 'dim_key' for slice dim")
            elif dim == 3:
                result_ext.time_dim = dim_key
            else:
                result_ext.vector_dim = dim_key
        result_hdr.extensions.append(result_ext)
        return NiftiWrapper(result_nii)


"""Provide a 'slight' normalization layer on top off DICOM meta data in a DcmMetaExtension
"""
import re
from dataclasses import dataclass
from typing import Optional, List, Callable, Set, Dict, Union, Tuple, Any

import pint

from .dcmmeta import DcmMetaExtension
from .convert import dt_to_datetime


ureg = pint.UnitRegistry()


@dataclass(frozen=True)
class SrcAttr:
    """A source attribute used to produce normalized attributes"""
    name: Union[str, Tuple[str, ...]]

    units: Optional[pint.Quantity] = None

    def __post_init__(self):
        if isinstance(self.name, str):
            object.__setattr__(self, "name", (self.name,))

    def get_source(self, meta: DcmMetaExtension):
        for name in self.name:
            cls = meta.get_classification(name)
            if cls is not None:
                return name


# Define some common source attributes
SRCS = {
    "AcquisitionDateTime" : SrcAttr(
        ("FrameReferenceDateTime", "FrameAcquisitionDateTime", "AcquisitionDateTime")
    ),
    "MosaicRefAcqTimes" : SrcAttr(
        ("CsaImage.MosaicRefAcqTimes", "SIEMENS_MR_HEADER.MosaicRefAcqTimes")
    ),
    "FlipAngle" : SrcAttr("FlipAngle", ureg.degrees),
    "RepetitionTime" : SrcAttr("RepetitionTime", ureg.milliseconds),
    "EchoTime" : SrcAttr(("EffectiveEchoTime", "EchoTime"), ureg.milliseconds),
    "InversionTime" : SrcAttr(("InversionTime", "InversionTimes"), ureg.milliseconds),
    "DiffusionBValue" : SrcAttr(
        (
            "DiffusionBValue", 
            "SIEMENS_MR_HEADER.0XC", 
            "SIEMENS_MR_HEADER.B_value", 
            "CsaImage.B_value",
            "GEMS_PARM_01.0X39",
            "PHILIPS_IMAGING_DD_001.0X3",
        ),
        ureg.seconds / ureg.mm ** 2,
    ),
}


@dataclass(frozen=True)
class SndAttr:
    """A normalized attribute we can extract from the meta data"""
    name: str

    units: Optional[pint.Quantity] = None

    short_name: Optional[str] = None

    def convert_single_source(self, src: SrcAttr, val):
        """Handle simple case of single source with potential unit conversion"""
        if self.units is None or src.units is None:
            return val
        if not isinstance(val, (float, int)):
            return [self.units.m_from(x * src.units) for x in val]
        return self.units.m_from(val * src.units)


# Define the normalized attributes
SND_ATTRS = [
    SndAttr("AcquisitionDescription"),
    SndAttr("ImageType"),
    SndAttr("AcquisitionTimeStamp", ureg.seconds),
    SndAttr("AcquisitionTimeStampSources"),
    SndAttr("FlipAngle", ureg.degrees, "FA"),
    SndAttr("RepetitionTime", ureg.seconds, "TR"),
    SndAttr("EchoTime", ureg.seconds, "TE"),
    SndAttr("InversionTime", ureg.seconds, "TI"),
    SndAttr("DiffusionBValue", ureg.seconds / ureg.mm ** 2, "b-value"),
]
SND_ATTR_MAP = {x.name: x for x in SND_ATTRS}


# Define some source attributes we assume are needed and don't need to be specified
GLOBAL_SOURCES = ["Manufacturer", "Modality"]


@dataclass(frozen=True)
class MetaNorm:
    """Define conversion from source meta data into normalized meta data
    
    Default `convert` function just returns the first source attribute that exists, 
    adjusting the units if needed.
    """
    out: Union[Union[SndAttr, str], Tuple[Union[SndAttr, str], ...]]

    sources: Optional[Tuple[Union[str, SrcAttr], ...]] = None

    convert: Optional[Callable[[Tuple[SrcAttr], DcmMetaExtension], Tuple[Any, ...]]] = None

    @property
    def n_out(self):
        return self._n_out

    def __post_init__(self):
        out = self.out
        n_out = 1
        if not isinstance(out, SndAttr):
            if isinstance(out, str):
                out = SND_ATTR_MAP[out] if out in SND_ATTR_MAP else SndAttr(out)
            else:
                n_out = len(out)
                out = tuple(SndAttr(x) if isinstance(x, str) else x for x in out)
            object.__setattr__(self, 'out', out)
        srcs = self.sources
        if srcs is None:
            srcs = [out.name] if n_out == 1 else [x.name for x in out]
        srcs = [SRCS.get(x, x) for x in srcs]
        object.__setattr__(
            self, 
            'sources', 
            tuple(SrcAttr(x) if isinstance(x, str) else x for x in srcs)
        )
        if self.convert is None:
            if n_out > 1:
                raise ValueError("No default 'convert' supports multiple 'out'")
            object.__setattr__(self, 'convert', self._default_convert)
        object.__setattr__(self, '_n_out', n_out)

    def _default_convert(self, meta: DcmMetaExtension):
        for src in self.sources:
            src_attr = src.get_source(meta)
            if src_attr is not None:
                val, cls = meta.get_values_and_class(src_attr)
                val = self.out.convert_single_source(src, val)
                return (val, cls)
        return (None, None)


@dataclass(frozen=True)
class ParseGroup:
    """Define parsing for group of related attributes"""
    attr_norms: Tuple[Union[str, MetaNorm], ...]

    applies: Optional[Callable] = None

    applies_sources: Optional[List[str]] = None
    
    def __post_init__(self):
        object.__setattr__(
            self, 
            'attr_norms', 
            tuple(MetaNorm(x) if isinstance(x, str) else x for x in self.attr_norms)
        )

    @property
    def sources(self) -> Set:
        res = set()
        if self.applies_sources:
            for src in self.applies_sources:
                res.add(src)
        for norm in self.attr_norms:
            for src in norm.sources:
                res.add(src)
        return res


def get_acq_descr(meta: DcmMetaExtension):
    gmeta = meta.get_class_dict(("global", "const"))
    srcs = ("SeriesDescription", "ProtocolName")
    mfgr = gmeta.get("Manufacturer")
    if mfgr is not None and re.search('Siemens', mfgr, re.IGNORECASE):
        srcs = ("ProtocolName", "SeriesDescription")
    for src in srcs:
        res = gmeta.get(src)
        if res is not None:
            return (res, ("global", "const"))
    return (None, None)


def get_image_type(meta: DcmMetaExtension):
    gmeta = meta.get_class_dict(("global", "const"))
    res = gmeta.get("ImageType", [])
    for add_type in gmeta.get("SIEMENS_MR_SDI_02.0X75", []):
        if add_type not in res:
            res.append(add_type)
    if not res:
        return (None, None)
    return (res, ("global", "const"))


def get_acq_ts(meta: DcmMetaExtension):
    srcs = []
    # The "FrameReferenceDateTime" is the ideal source, as it should correspond to the
    # center of k-space
    for key in ("FrameReferenceDateTime", "FrameAcquisitionDateTime" "AcquisitionDateTime"):
        acq_dt, cls = meta.get_values_and_class(key)
        if acq_dt is not None:
            srcs.append(key)
            if isinstance(acq_dt, str):
                acq_dt = [acq_dt]
            break
    else:
        for key in ("AcquisitionTime",):
            acq_tm, cls = meta.get_values_and_class(key)
            if acq_tm is not None:
                if isinstance(acq_tm, str):
                    acq_tm = [acq_tm]
                srcs.append(key)
                break
        else:
            return ((None, None), (None, None))
        # TODO: If AcquisitionDate is define can we skip doing a wrap check?
        needs_wrp_check = False
        for key in ("AcquisitionDate", "SeriesDate"):
            acq_date = meta.get_values(key)
            if acq_date is not None:
                srcs.append(key)
                break
        else:
            acq_date = "19000101"
        if isinstance(acq_date, str):
            needs_wrp_check = True
            acq_dt = [acq_date + tm for tm in acq_tm]
        else:
            for dt, tm in zip(acq_date, acq_tm):
                acq_dt = [dt + tm]
        # TODO: Implement heurstic check if times span more than one day
    acq_ts = [dt_to_datetime(dt).timestamp() for dt in acq_dt]
    for key in ("CsaImage.MosaicRefAcqTimes", "SIEMENS_MR_HEADER.MosaicRefAcqTimes"):
        vals = meta.get_values(key)
        if vals is not None:
            assert len(acq_ts) == 1
            base_ts = acq_ts[0]
            acq_ts = [base_ts + (x / 1000.0) for x in vals]
            cls = ("global", "slices")
            srcs.append(key)
            break
    if cls == ("global", "const"):
        acq_ts = acq_ts[0]
    return ((acq_ts, cls), (srcs, ("global", "const")))
    


GEN_ATTRS = ParseGroup(
    (
        MetaNorm(
            "AcquisitionDescription", ("SeriesDescription", "ProtocolName"), get_acq_descr
        ),
        MetaNorm(
            ("AcquisitionTimeStamp", "AcquisitionTimeStampSources"), 
            ("AcquisitionTime", "AcquisitionDate", "AcquisitionDateTime", "MosaicRefAcqTimes"),
            get_acq_ts,
        ),
        MetaNorm("ImageType", ("ImageType", "SIEMENS_MR_SDI_02.OX75"))
    )
)


MR_ATTRS = ParseGroup(
    (
        "FlipAngle",
        "RepetitionTime",
        "EchoTime",
        "InversionTime",
        "DiffusionBValue",
    ),
    lambda meta: meta.get_values("Modality") == "MR"
)


SND_GROUPS = (GEN_ATTRS, MR_ATTRS)


def get_snd_sources(groups=SND_GROUPS) -> Set:
    """Get set containing all the needed source meta data keys"""
    res = set(GLOBAL_SOURCES)
    for grp in groups:
        for src in grp.sources:
            res |= set(src.name)
    return res


SND_VERS = "20241118"


def gen_snd_meta(meta: DcmMetaExtension, groups: Tuple[ParseGroup, ...] = SND_GROUPS):
    """Generate the normalized meta data
    
    At least for now we assume `meta` comes from a single DICOM file.
    """
    for grp in groups:
        if grp.applies is not None and not grp.applies(meta):
            continue
        for norm in grp.attr_norms:
            res = norm.convert(meta)
            if norm.n_out == 1:
                val, cls = res
                if val is not None:
                    yield (cls, norm.out.name, val)
            else:
                for out_idx, (val, cls) in enumerate(res):
                    if val is not None:
                        if val is not None:
                            yield (cls, norm.out[out_idx].name, val)
    yield (("global", "const"), "__version__", SND_VERS)


def inject(
    meta: DcmMetaExtension, 
    prefix: str = "SND.", 
    groups: Tuple[ParseGroup, ...] = SND_GROUPS
):
    """Inject normalized meta data into the DcmMetaExtension"""
    for cls, key, val in gen_snd_meta(meta, groups):
        meta.get_class_dict(cls)[f"{prefix}{key}"] = val



"""
Various utils and helpers
"""

import sys, string, traceback, pdb

# Python 2 / 3 compatibility
# Ideally should just use six module
PY2 = sys.version_info[0] < 3

unicode_str = unicode if PY2 else str
byte_str = str if PY2 else bytes
str_types = (unicode_str, byte_str)
iteritems = (lambda d: d.iteritems()) if PY2 else lambda d: d.items()
ascii_letters = string.letters if PY2 else string.ascii_letters


def pdb_except_hook(type, value, tb):
    """Function that can be used as excepthook to start PDB"""
    traceback.print_exception(type, value, tb)
    print()
    pdb.post_mortem(tb)



"""
Module providing tests for dcmstack
"""
import sys
from os import path

test_dir = path.dirname(__file__)
src_dir = path.normpath(path.join(test_dir, path.pardir, 'src'))

if sys.path[0] != src_dir:
    sys.path.insert(0, src_dir)



import sys, shutil, json
from os import path as opath
from tempfile import mkdtemp
from glob import glob

import numpy as np

from . import test_dir, src_dir

from dcmstack import dcmstack_cli, nitool_cli

test_data = {}

def setup_module():
    input_dir = mkdtemp(prefix='dcmstack_cli_test_in')
    data_dir = opath.join(test_dir,
                          'data',
                          'dcmstack',
                          '2D_16Echo_qT2')
    fns = ('TE_20_SlcPos_-33.707626341697.dcm',
           'TE_20_SlcPos_-23.207628249046.dcm',
           'TE_40_SlcPos_-33.707626341697.dcm',
           'TE_40_SlcPos_-23.207628249046.dcm',
           )
    for fn in fns:
        shutil.copyfile(opath.join(data_dir, fn), opath.join(input_dir, fn))
    test_data['base_dir'] = input_dir


def teardown_module():
    shutil.rmtree(test_data['base_dir'])
    del test_data['base_dir']


def make_niftis(out_dir, extra_args=tuple()):
    '''Helper function runs dcmstack and returns any Nifti files  
    '''
    assert len(glob(opath.join(out_dir, '*.nii*'))) == 0
    args = ['dcmstack', test_data['base_dir'], '--strict', '--dest-dir', out_dir]
    args.extend(extra_args)
    dcmstack_cli.main(args)
    return glob(opath.join(out_dir, '*.nii*'))


class CliTest(object):
    def setup_method(self):
        self.out_dir = mkdtemp(prefix='dcmstack_cli_test_out')

    def teardown_method(self):
        shutil.rmtree(self.out_dir)


class TestDcmstackCli(CliTest):
    def test_basic(self):
        nii_paths = make_niftis(self.out_dir)
        assert len(nii_paths) == 1

    def test_embed(self):
        nii_paths = make_niftis(self.out_dir, ['--embed'])
        assert len(nii_paths) == 1


class TestNitoolCli(CliTest):
    def test_basic(self):
        nii_path = make_niftis(self.out_dir, ['--embed'])[0]
        json_path = opath.join(self.out_dir, 'meta.json')
        nitool_cli.main(['nitool', 'dump', nii_path, json_path])
        with open(json_path) as fp:
            meta = json.load(fp)
        print(json.dumps(meta, indent=4))
        assert('dcmmeta_version' in meta)
        assert('dcmmeta_affine' in meta)
        assert('dcmmeta_reorient_transform' in meta)
        assert meta['dcmmeta_shape'] == [192, 192, 2, 2]
        assert meta['dcmmeta_slice_dim'] ==  2
        assert meta['global']['const']['Rows'] == 192
        assert meta['global']['const']['Columns'] == 192
        assert(np.allclose(meta['time']['samples']['EchoTime'], [20.0, 40.0]))
